"""External methods for Tracker-related ZClasses and Products."""

#import Products.TrackerBase.TrackerMisc # DEBUG
#reload(Products.TrackerBase.TrackerMisc) # DEBUG

from Products.TrackerBase.TrackerMisc import *

import OFS.Image

import sys, time

TRACKER_DATA_VERSION = 1
ISSUE_DATA_VERSION = 1
ITEM_DATA_VERSION = 1

def addTracker(self, id, REQUEST=None):
    """ """
    # self = Factory - actually, FactoryDispatcher
    # Destination() = the folder in which the obj is being created.


    if not id:
        raise ValueError, "You must specify an id."
    inst = self.Tracker(id)

    dest = self.Destination()
    dest._setObject(id, inst)
    tracker = getattr(dest, id)
    tracker.DATA_VERSION = TRACKER_DATA_VERSION

    user = (REQUEST and `REQUEST.AUTHENTICATED_USER`) or ''
    tracker.manage_addLocalRoles(user, ['TrackerOwner'])

    for role in unique(ROLE_CATEGORIES['open']
                       + ROLE_CATEGORIES['dedicated']):
        if not tracker._has_user_defined_role(role):
            tracker._addRole(role)

    setTrackerRules(tracker)

    tracker.smtphost = socket.gethostname()
    tracker.smtpport = 25

    for ps in tracker.propertysheets:
        ps.manage_changeProperties(REQUEST)

    tracker.supporters = []
    tracker.clients = []
    tracker.peers = []
    tracker.open_subscribe = tracker.default_open_subscribe

    trackerInitCatalog(tracker)

    if REQUEST is not None:
        return REQUEST['RESPONSE'].redirect(self.DestinationURL()
                                            + '/' + id
                                            + '/configBasicForm')

def setTrackerConfig(self, REQUEST):
    """Process fundamental configuration activities.

    If request has a 'nextpage' setting, we redirect there, otherwise we go 
    to the tracker proper."""

    # XXX Will this be needed for site search stuff?
    #self.unindex_object()
    if REQUEST.get('config_form', '') == 'basic':
        setTrackerConfigBasic(self, REQUEST)

    elif REQUEST.get('config_form', '') == 'traitvalues':
        setTrackerConfigTraitValues(self, REQUEST)

    elif REQUEST.get('config_form', '') == 'assignments':
        setTrackerConfigAssignments(self, REQUEST)

    elif REQUEST.get('config_form', '') == 'reconcile':
        reconcile_tracker(self, REQUEST)

    else:
        raise 'Bad Request', "Unrecognized tracker configuration activity."

    for ps in self.propertysheets:
        ps.manage_changeProperties(REQUEST)

    # XXX Will this be needed for site search stuff?
    #self.index_object()

    REQUEST.set('manage_tabs_message', 'Changed!')
    REQUEST["RESPONSE"].redirect(REQUEST.get('nextpage',
                                             self.absolute_url()))

def reconcile_tracker(tracker, REQUEST):
    """Iterate through elements changing to accomodate new developments."""
    roles = unique(ROLE_CATEGORIES['open'] + ROLE_CATEGORIES['dedicated'])

    do_structure = REQUEST.get('reconcileStructure', None)
    do_traitsUpdate = REQUEST.get('updateTraits', None)
    do_catalog = REQUEST.get('reconcileCatalog', None)

    if do_structure:
        setTrackerRules(tracker)
    if do_traitsUpdate or do_structure:
        for issue in tracker.values():
            if do_structure:
                audit_vars('StatusVars', issue, tracker, REQUEST)
                for role in roles:
                    if not tracker._has_user_defined_role(role):
                        tracker._addRole(role)
            if do_traitsUpdate:
                setTraitValues(issue, tracker, extend=1)

    if do_catalog:
        trackerInitCatalog(tracker)

def setTrackerConfigBasic(self, REQUEST):
    """Do setTrackerConfig operations for basic config form."""

    self.did_basic = 1

    traits = REQUEST.get('traits', [])
    traits = self.traits = unique(map(string.strip, traits), noblanks=1)
    REQUEST.set('traits', traits)

    # Tracker access mode
    mode = REQUEST.get('access_mode', 'open')
    if mode == 'dedicated' and not hasattr(self, 'clients'):
        self.clients = []
    if mode != self.access_mode:
        if mode == 'open':
            # Permissive
            self.manage_permission('View', roles=ROLE_CATEGORIES['open'],
                                   acquire=1)
        else:
            # Restrictive
            self.manage_permission('View', roles=ROLE_CATEGORIES['dedicated'],
                                   acquire=0)
        self.access_mode = mode

    # tracker.traitsStruct: {range_nm: (val_nm, [vals, ...])}
    # Adjust or create user specified traits data structure
    traitsStruct = {}
    use_current = 0
    if hasattr(self, 'traitsStruct'):
        # Reconcile with prior version of traits struct.
        traitsStruct.update(self.traitsStruct)
        traitsRange = map(lambda x: x + "_range", traits)
        # Cull defunct traits:
        for i in traitsStruct.keys():
            if i not in traitsRange:
                del traitsStruct[i]
    for val_nm in traits:
        range_nm = val_nm + "_range"
        if traitsStruct.has_key(range_nm):
            continue
        elif hasattr(self, range_nm):
            traitsStruct[range_nm] = (val_nm, getattr(self, range_nm))
        else:
            traitsStruct[range_nm] = (val_nm, [])
    self.traitsStruct = traitsStruct

    if not hasattr(self, 'pickers'):
        pickers = {}
    else:
        pickers = self.pickers
        for i in pickers.keys():
            if i not in traits + [' __requesters']:
                del pickers[i]
    self.pickers = pickers

def setTrackerConfigTraitValues(self, REQUEST):
    """Adjust traitsStructure according to trait/values config form."""

    self.did_traitvalues = 1

    if self.access_mode == 'dedicated':
        self.support_level = REQUEST.get('support_level', self.support_level)
        # Massage date fields into cohesive date settings.
        for i in ('start', 'end'):
            mo, day, yr = (REQUEST.get('tickets_%s_month' % i, 0),
                           REQUEST.get('tickets_%s_day' % i, 0),
                           REQUEST.get('tickets_%s_year' % i, 0))
            dstr = "%s %s, %s" % (mo, day, yr)
            if not (mo and day and yr):
                raise ValueError, "Invalid date spec: %s" % dstr
            REQUEST.set('tickets_%s_date' % i, dstr)

    # tracker.assignments: {supporter: {trait: [vals, ...]}}
    sa = {}  
    if hasattr(self, 'assignments'):
        sa.update(self.assignments)
    # tracker.supporters: [supporter, ...]
    if hasattr(self, 'supporters'):
        were_supporters = filter(None, self.supporters)
    else:
        were_supporters = []
    are_supporters = unique(REQUEST.get('supporters', []), noblanks=1)
    are_supporters.sort
    REQUEST.set('supporters', are_supporters) # Sans blanks.
    self.supporters = are_supporters
    added, defunct = diffs(are_supporters, were_supporters)
    for i in defunct:
        if sa.has_key(i):
            del sa[i]
        clear_local_role('Supporter', self, user=i)
    for i in added:
        self.manage_addLocalRoles(i, ['Supporter'])

    traits = self.traits

    # Adjust or create supporter/trait assignments data structure.
    for s in are_supporters:
        d = sa.get(s, {})
        for t in traits:
            if not d.has_key(t):
                d[t] = []
        sa[s] = d
    self.assignments = sa

    # tracker.clients: [supporter, ...]
    if hasattr(self, 'clients'):
        were_clients = filter(None, self.clients)
    else:
        self.clients = []
        were_clients = []
    are_clients = unique(REQUEST.get('clients', []), noblanks=1)
    self.clients = are_clients
    added, defunct = diffs(are_clients, were_clients)
    for i in defunct:
        clear_local_role('TrackerClient', self, user=i)
    for i in added:
        self.manage_addLocalRoles(i, ['TrackerClient'])

    traitsStruct = self.traitsStruct
    for val_nm in self.traits:
        range_nm = val_nm + "_range"
        newvals = REQUEST.get(range_nm, None)
        if newvals is not None:
            newvals = unique(newvals, noblanks=1)
            traitsStruct[range_nm] = (val_nm, newvals)
    self.traitsStruct = traitsStruct

    pickers = self.pickers
    for i in traits:
        if REQUEST.get(i + "__picker", None):
            pickers[i] = 1
        elif pickers.has_key(i):
            del pickers[i]
    if REQUEST.get(" __requesters", None):
        pickers[" __requesters"] = 1
    elif pickers.has_key(" __requesters"):
        del pickers[" __requesters"]
    self.pickers = pickers

    # Check for supporters or clients lacking email and override
    # REQUEST nextpage setting if necessary.
    subjs = self.supporters
    if (hasattr(self, 'clients')
        and (self.access_mode == 'dedicated') and self.clients):
        subjs = subjs + self.clients
    for i in subjs:
        u = trackerGetUser(self, i)
        if not hasattr(u, 'email') or not u.email:
            dnp = REQUEST.get('nextpage', '')
            if dnp:
                dnp = '?deferrednextpage=' + dnp
            REQUEST.set('nextpage', 'configMemberSettings' + dnp)
            REQUEST.set('MissingEmailMsg', 1)
            return

def setTrackerConfigAssignments(self, REQUEST):
    """Do setTrackerConfig operations for supporter assignments config form."""

    self.did_assignments = 1

    pickers = self.pickers
    assignments, supporters = self.assignments, self.supporters

    for s in assignments.keys():
        # Cull obsolete assignment supporters
        if s not in supporters:
            del assignments[s]

    for s in supporters:
        sa = assignments.get(s, {})
        # Cull obsolete assignment categories
        for k in sa.keys():
            if not pickers.has_key(k):
                del sa[k]
        for t in pickers.keys():
            curr = REQUEST.get(s + "__" + t, [])
            if curr and (curr[0] == " __INVERT"):
                # Inverse - represented by an empty tuple.
                curr[0] = ()
            sa[t] = curr

    self.assignments = assignments
 
def addTrackerIssue(self, REQUEST=None):
    """Do permission setting and item allocations for new issue.

    Create a new issue and use TrackerBase.add_issue() to add it to the
    tracker.  Assign the user who submitted the issue with the Request role,
    initialize issue properties like privacy, traits, etc, and catalog the new
    issue.  Create a new item using the new issue's .addIssueItem() method,
    and return a RESPONSE at the new issue summary. """

    REQUEST.set('date', DateTime())

    __traceback_info__=('addTrackerIssue', `self`)
    if hasattr(self, 'Destination'):
        # We're in a factory.
        if REQUEST:
            tracker = REQUEST.PARENTS[0]
        else:
            tracker = self.Destination()
    else:
        tracker = self
    issue = tracker.TrackerIssue(REQUEST)
    issueid = tracker.add_issue(issue)
    issue.DATA_VERSION = ISSUE_DATA_VERSION
    issue = tracker[issueid]
    requester = `REQUEST.AUTHENTICATED_USER`
    issue.requester = requester
    issue._addRole('Requester')
    issue.manage_setLocalRoles(requester, ['Requester'])

    if type(tracker.open_subscribe) == type(''):
        if string.lower(tracker.open_subscribe) == 'on':
            issue.open_subscribe = 1
        else:
            issue.open_subscribe = 0
    else:
        issue.open_subscribe = int(tracker.open_subscribe)
    issue.subscribers = []
    issue.references = []
    issue.diverted = ''
    issue.reframed = ''

    for ps in issue.propertysheets:
        ps.manage_changeProperties(REQUEST)

    if REQUEST.get('confidential', 0):
        issue.private = 1

    # Derive issue properties and audit trail from tracker properties:
##     # Initialize the issue property sheet:
##     __traceback_info__ = (`issue.propertysheets`)
##     currentstate = PropertySheet('CurrentState')
##     for issueid, value, type in properties:
##         currentstate.manage_addProperty(issueid, value, type)
##     currentstate = FixedSchema('CurrentState', currentstate)
##      issue.propertysheets.manage_addPropertySheet('CurrentState',
##                                                   currentstate)

    issue.auditedAttrs = audit_vars('StatusVars', issue, tracker, REQUEST)
    setTraitValues(issue, tracker, REQUEST)
    REQUEST.set('from', requester)
    # This will be done as a 'recatalogElement' in addIssueItem:
    catalogElement(issue)
    addIssueItem(issue, msgrole='request', REQUEST=REQUEST)

    issue.set_privacy(issue.private, asissue=1)

    if REQUEST is not None:
        return REQUEST['RESPONSE'].redirect(REQUEST.URL1 + '/' + str(issueid))

def setTraitValues(issue, tracker, REQUEST=None, extend=0):
    """Populate issue with trait/values, establishing fresh or extending."""
    traitVals = []                 # [(traitname, value), ...]
    traitdict = {}
    origTraitValsDict = {}
    traitsStruct = tracker.traitsStruct
    if hasattr(issue, 'traitVals'):
        for t, v in issue.traitVals:
            origTraitValsDict[t] = v
    for i in tracker.traits:
        if REQUEST is None: val = None
        else: val = REQUEST.get(i + "__trait", None)
        if val is not None:
            traitVals.append(i, val)
            traitdict[i] = val
        elif origTraitValsDict.has_key(i):
            traitVals.append(i, origTraitValsDict[i])
            traitdict[i] = origTraitValsDict[i]
        else:
            traitdict[i] = ''
            if extend:
                vals = traitsStruct[i + "_range"][1]
                if vals: traitVals.append(i, vals[0])
                else: traitVals.append(i, '')

    if not hasattr(issue, 'traitVals') or issue.traitVals != traitVals:
        issue.traitVals = traitVals
##    # XXX The following audits, but seems to cause authentication problems?
##    if not hasattr(issue, 'traitVals') or issue.traitVals != traitVals:
##        userid = REQUEST.AUTHENTICATED_USER.getUserName()
##        if not hasattr(issue, 'traitVals'):
##            issue.audited_state.set('setTraitValues', 'new', userid)
##        else:
##            # XXX May eventually be audit useful to register change details.
##            issue.audited_state.set('setTraitValues', 'change', userid)
##        issue.traitVals = traitVals

    # Try to adopt initial priority settings from related trait settings:
    for i in traitdict.keys():
        if string.lower(i) in ['priority', 'urgency', 'importance', 'need']:
            val = string.lower(traitdict[i])
            if val in tracker.priority_range:
                issue.priority = val
                break

def reattrs(self, REQUEST=None):
    """Temporary - rederive audited attrs for an issue."""
    tracker, issue, item = get_elements(self)
    issue.auditedAttrs = audit_vars('StatusVars', issue, tracker, REQUEST)
    if REQUEST is not None:
        return REQUEST['RESPONSE'].redirect(REQUEST.URL1
                                            + '/' + 'editStateYikes')

def audit_vars(varrange, issue, tracker, REQUEST):
    """Package the structure of state var settings in an issue."""

    # A list of the attributes names and a dictionary going from
    # attibute names and the range of values are kept in a dictionary
    # on the issue's auditedAttrs attribute.
    if not hasattr(issue, varrange):
        raise AttributeError, ("'%s' var range property missing."
                               % varrange)
    dict = {}
    seq = []
    # Derive the data structures from the named :lines property.
    additions = []
    for var_nm in getattr(issue, varrange):
        if not var_nm:
            continue
        range_nm = var_nm + "_range"

        if hasattr(issue, range_nm):
            vals = getattr(issue, range_nm)
        elif hasattr(tracker, range_nm):
            vals = getattr(tracker, range_nm)
        else:
            raise AttributeError, ("State property '%s' not found"
                                   " (see VarInstances property '%s')"
                                   % (varrange, range_nm))
        vals = filter(None, vals)
        seq.append(var_nm)
        dict[var_nm] = vals

        # Institute the attribute and initial value on the issue:
        if hasattr(issue, var_nm):
            curval = getattr(issue, var_nm)
            if type(curval) == type([]):
                curval = curval[0]
        else:
            curval = vals[0]
        setattr(issue, var_nm, curval)
        # Register the audit trail:
        additions.append((var_nm, curval))
    issue.add_state(additions, REQUEST)
    if not hasattr(issue, 'stage'):
        issue.stage = 'pending'
    if not(hasattr(issue, 'auditedAttrs')):
        return {varrange: (seq, dict)}
    else:
        issue.auditedAttrs[varrange] = (seq, dict)
        return issue.auditedAttrs

def addIssueItem(self, msgrole='correspond', REQUEST=None):
    """Create a new item.  Raises an informative redirect for invalid input.

    Populate the issue with a freshly minted item, setting the item's action
    from the msgrole setting and the description, from, date, etc from the
    REQUEST, and doing a lot of incidental business: 

    - Create a new item
    - Add the item to the issue
    - Register the item's action according to the msgrole.
    - Set some other item metadata according to the REQUEST and incidental
      data (time, etc.) 
    - Do divert, if any
    - Set the issue and item stage using setIssueStage(), IF valid.
    - Stash user's new signature and/or alternate email addr if specified
    - if coming from another item, hookup the mutual "replies"/"in-reply-to"
    - set item privacy
    - append any item reference objects to issue references
    - execute rules
    - catalog the new item and recatalog the issue
    - send email notification
    """

    signature = REQUEST.get('signature', '')
    if signature:
        description = REQUEST.get('description', '') + '\n\n' + signature
        REQUEST.set('description', description)
    if not (string.strip(REQUEST.get('description', ''))
            or string.strip(REQUEST.get('title', ''))):
        errmsg = ('<h3><font color="red">Issue must have a non-empty'
                  ' title or description.</font></h3>'
                  'Hit back to return to the form.')
        raise 'Redirect', errmsg

    REQUEST.set('date', DateTime())

    attachments, errmsg = snarfAttachments(REQUEST)
    if errmsg:
        errmsg = ('<h3><font color="red">'
                  + errmsg
                  + '</font></h3>'
                  + "Hit back to return to the form.")
        raise 'Redirect', errmsg

    # Stage changes (watch out for privacy conflict)

    tracker, issue, item = get_elements(self)
    userobj = REQUEST.AUTHENTICATED_USER
    user = userobj.getUserName()
    (itemid, item) = issue.add_item(issue.TrackerItem(REQUEST), user)
    issue.DATA_VERSION = ISSUE_DATA_VERSION

    no_post = REQUEST.get('no_post', 0)

    if REQUEST.get('private', 0) and itemid != 1:
        item.private = 1

    wasstage = issue.stage

    item.msgrole = msgrole
    if msgrole == 'request':
        setIssueStage(item, 'pending', REQUEST, no_promote=no_post)
    elif msgrole == 'accept':
        setIssueStage(item, 'accepted', REQUEST, no_promote=no_post)
    elif msgrole == 'resolve':
        setIssueStage(item, 'resolved', REQUEST, no_promote=no_post)
    elif msgrole == 'reject':
        setIssueStage(item, 'rejected', REQUEST, no_promote=no_post)
    elif msgrole == 'defer':
        setIssueStage(item, 'deferred', REQUEST, no_promote=no_post)
    elif msgrole == 'divert':
        errmsg = divert_issue(issue, item, REQUEST)
        if errmsg:
            get_transaction().abort()
            raise 'Redirect', errmsg
        setIssueStage(item, 'diverted', REQUEST, no_promote=no_post)

    # The following msgroles don't change the stage.
    elif msgrole == 'subscribe':
        issue.manage_addLocalRoles(user, ['IssueOwner'])
    elif msgrole == 'resign':
        if len(issueOwners(issue)) <= 1:
            errmsg = ('The issue must have more than one owner'
                      ' for you to resign')
            raise 'Redirect', errmsg
        clear_local_role('IssueOwner', issue, user=user)
    elif msgrole != 'correspond' and not no_post:
        if item.private:
                raise 'Bad Request', ("Action summary items cannot be private."
                                      '\n - either select "Don\'t post" or'
                                      " submit a non-action message.")
        issue.promotes(item.id, user)
    elif msgrole == 'reframe':
        # Promote an alternate presentation of the request description.
        # Nothing to do til after properties gleaned.
        pass

    afn = REQUEST.get('alt_full_name', ())
    if afn:
        if afn != ((hasattr(userobj, 'full_name') and userobj.full_name)
                   or ''):
            if type(issue.alt_full_names) == type({}):
                issue.alt_full_names[user] = afn
            else:
                issue.alt_full_names = {user: afn}
        elif ((type(issue.alt_full_names) == type({}))
              and issue.alt_full_names.has_key(user)):
            # Null so we track member changes of their central info.
            del issue.alt_full_names[user]
    ae = REQUEST.get('alt_email', ())
    if ae:
        if (not '@' in ae) or (' ' in string.strip(ae)):
            raise ValueError, ("<font color='red'>Invalid alternate email"
                               " address '%s'.  Please return and"
                               " correct it.</font>"
                               % ae)
        if ae != ((hasattr(userobj, 'email') and userobj.email) or ''):
            if type(issue.alt_emails) == type({}):
                issue.alt_emails[user] = ae
            else:
                issue.alt_emails = {user: ae}
        elif ((type(issue.alt_emails) == type({}))
              and issue.alt_emails.has_key(user)):
            # Null so we track member changes of their central info.
            del issue.alt_emails[user]

    if REQUEST.get('alt_full_name', None):
        userobj.full_name = REQUEST.get('alt_full_name')
    if REQUEST.get('alt_email', None):
        userobj.email = REQUEST.get('alt_email')

    if REQUEST.get('retainsig', ''):
        trackerpref(pref='signature', value=signature, REQUEST=REQUEST)

    for ps in item.propertysheets:
        ps.manage_changeProperties(REQUEST)
                      
    origin = REQUEST.get("in_reply_to", '')
    if origin:
        try:
            origin_item = issue[origin]
            psheet = origin_item.propertysheets.Configuration
            replyprop = filter(None, psheet.getProperty('replies'))
            psheet.manage_changeProperties({'replies':
                                            replyprop + [itemid]})
        except IndexError:
            pass

    # Add any references to the issue-level references value.
    if filter(None, item.references):
        changed = 0
        for r in item.references:
            if r not in issue.references:
                issue.references.append(r)
                changed = 1
        if changed:
            issue.references = issue.references

    # Attachments stuff:
    for aid, origtitle, title, file in attachments:
        if not origtitle:
            title = aid                 # Avoid showing the file whole path.
        item.manage_addFile(aid, file, title)

    if int(itemid) in issue.promotes():
        item.set_privacy(issue.private, asissue=1)
        item.private = 0
    else:
        item.set_privacy(item.private, asissue=0)

    if msgrole == 'reframe':
        # This has to come late enough to have gleaned the id, etc.
        issue.reframed = item.id
        issue.description = item.description
        issue.title = item.title

    issue.execute_rules()

    catalogElement(item)
    recatalogElement(issue)

    if not no_post:
        msgid = '[t1[%s](%s)]' % (issue.id, itemid)
        send_posting(item, REQUEST.AUTHENTICATED_USER, wasstage,
                     REQUEST=REQUEST)

    if REQUEST is not None:
        if msgrole == 'reframe':
            respred = REQUEST['RESPONSE'].redirect
            msg = urllib.quote('manage_tabs_message=<font color="green">'
                               'Adjust issue metadata.</font>')
            return respred(REQUEST.URL1 + '/editStateYikes?' + msg)
        if msgrole in ['resolve', 'defer', 'reject']:
            respred = REQUEST['RESPONSE'].redirect
            msg = urllib.quote('manage_tabs_message=<font color="green">'
                               'Issue %s %s done.</font>' % (issue.id, 
                                                             msgrole))
            return respred(tracker.absolute_url() + '?' + msg)
        else:
            return REQUEST['RESPONSE'].redirect(REQUEST.URL1)

def divert_issue(issue, item, REQUEST):
    """Subsume an issue in another, according to REQUEST references.

    We return an error message on failure, else None.

    If the divert target issue doesn't exist, or if the parties to the
    diverted issue are unable to see the target issue, the divert fails.

    If the target issue is active (stage is accepted or pending) then the
    target issue parties are notified of the divert.

    The parties to the diverted issue are subscribed to the target issue,
    mutual references in both issues are registered, and any alternate id
    stuff (full names, emails) are ported from the old to the new issue."""

    ## XXX Currently we fail to port anonymous or superuser requesters and
    ##     subscribers to the target issue of the divert.  This is a bug, but 
    ##     it's a hassle to fix and i'm a bit undermotivated to cater to
    ##     anonymous/superuser submitters.

    references = REQUEST.get("references", '')
    targetobj = None
    if references:
        references = filter(None, references)
        if not references:
            REQUEST.set('references', references)
        else:
            # Verify that the target is an issue.
            hrefs = hrefize(issue, references[0], REQUEST=REQUEST)[1]
            if hrefs and type(hrefs[0]) != type(""):
                targetobj = hrefs[0]
    if not targetobj:
        return ('<strong> <font color="red">Divert failed</font> - </strong>'
                '<br> <strong> You must specify the target issue as first'
                '     reference in the references entry box.</strong>\n'
                '<p> - use <em>(issue-number)[]<em>'
                '      or <em>(issue-number)[item-number]</em> format.</p>\n'
                '<p> Hit back to return to the form and designate the target'
                '    issue.</p>\n')
    tgt_tracker, tgt_issue, tgt_item = get_elements(targetobj)
    if tgt_issue.absolute_url() == issue.absolute_url():
        return ('<strong> <font color="red">Divert failed</font> - </strong>'
                '<br> <strong> You cannot divert an issue to itself. </strong>'
                '<p> Hit back to return to the form.</p>\n')

    # Ensure parties to diverted issue can view target.
    # XXX Eventually offer choices, simple override or:
    #     - endow those who can't view with privilege,
    #     - abandon those who can't view, 
    requester = issue.requester
    unpriv = ""
    numunpriv = 0
    for i in [requester] + issue.subscribers:
        if not(viewable(targetobj, userid=i)):
            unpriv = unpriv + i + " "
            numunpriv = numunpriv + 1
    if unpriv:
        return ('<strong><font color="red">Divert failed</font></strong> - '
                ' <p> Diverted issue participant%s <strong> %s </strong>'
                ' not privileged to view the target issue'
                ' %s.' 
                '</p> Hit back to return to the form and change your'
                '<br> target, or, if suitable, visit the target and change its'
                '<br> privacy or security settings to enable access.\n'
                % (((numunpriv > 1) and "s") or "",
                   unpriv,
                   elementEmblem(tgt_issue, complete=1, REQUEST=REQUEST)))

    target_emblem = '%s(%s)[]' % (tgt_tracker.abbrev, tgt_issue.id)
    diverted_emblem = '%s(%s)[]' % (issue.aq_parent.abbrev, issue.id)
    if tgt_issue.stage in ['accepted', 'pending']:
        d = DateTime()
        text = 'Subject: [%s] Subsuming issue %s\n' % (target_emblem,
                                                       diverted_emblem)
        text = text + ('From: Tracker System for <%s>\n' % 
                       tgt_tracker.adminaddr)
        text = text + 'To: Issue Participants <>\n'
        text = text + 'Date: %s\n' % d.rfc822()
        text = text + '\n'
        owners = issueOwners(issue)
        others = [issue.requester] + issue.subscribers
        sender = REQUEST.AUTHENTICATED_USER
        recipients = []
        if not tracker_role(item, ['superuser', 'Anonymous User'],
                            sender, user=sender):
            recipients.append(sender.getUserName())
        recipient_ids = unique(owners + others)
        recipients_str = ''
        for i in recipient_ids:
            recipients_str = recipients_str + ' ' + i
            memberobj = trackerGetUser(tgt_tracker, i)
            if memberobj and hasattr(memberobj, 'email') and memberobj.email:
                recipients.append(memberobj.email)
        text = text + ('Issue %s has been diverted to your issue %s.\n'
                       'Diverted issue: %s\n'
                       'Your issue: %s\n\n'
                       'This means that the issue will be subsumed in yours,\n'
                       'including the %s requesters and any subscribers:\n'
                       % (diverted_emblem, target_emblem,
                          issue.absolute_url(), tgt_issue.absolute_url(),
                          diverted_emblem))
        text = text + '    ' + recipients_str + '\n'
        if recipients:
            send_email(tgt_tracker, tgt_tracker.adminaddr, recipients, text)

    already_ids = [tgt_issue.requester] + tgt_issue.subscribers
    for i in [requester] + issue.subscribers:
        if i in already_ids:
            continue
        targetobj.subscribers.append(i)
    # Endow the requester with posting privileges in the target.
    tgt_issue.manage_addLocalRoles(requester, ['Requester'])
    # Register the diversion in the issue and the target issue.
    issue.diverted = elementEmblem(tgt_issue, nolink=1, complete=1,
                                   REQUEST=REQUEST)
    tgt_issue.references.append(elementEmblem(issue, nolink=1, complete=1,
                                              REQUEST=REQUEST))

    # Merge any alternate identification stuff from old issue into target:
    if issue.alt_emails:
        if not tgt_issue.alt_emails:
            tgt_issue.alt_emails = {}
        tgt_issue.alt_emails.update(issue.alt_emails)
    if issue.alt_full_names:
        if not tgt_issue.alt_full_names:
            tgt_issue.alt_full_names = {}
        tgt_issue.alt_full_names.update(issue.alt_full_names)
            
    tgt_issue.audited_state.set('subsume_issue', diverted_emblem,
                                `REQUEST.AUTHENTICATED_USER`)

def singlifyQuotes(str):
    return string.replace(str, '"', "'")

def send_posting(item, authuser, wasstage, sender=None, REQUEST=None):
    """Send item posting notices to the "right" recipients.

    For unowned issues, sends to relevant supporters if from requester, or to
    relevant supporters and requester if from supporter.  "Relevant"
    supporters are chosen according to trait and requester assignments.

    Similar for owned issues except "relevant supporter" means the, and only
    the, issue-owning supporter.

    A courtesy copy is always sent to the party sending the message."""
    # XXX We currently do not exclude originator - we may need a metoo option,
    #     and/or a form checkbox, so they can control this behavior.

    tracker, issue, garbage = get_elements(item)
    acl_users = item.acl_users
    capwords = string.capwords

    if sender is None:
        sender = tracker.adminaddr

    userid = `authuser`
    afn = issue.alt_full_names
    if afn and afn.has_key(userid):
        usernm = "%s (%s)" % (afn[userid], userid)
    elif hasattr(authuser, 'full_name') and authuser.full_name:
        usernm = authuser.full_name
    else:
        usernm = userid

    # Identify the requester, subscribers, and issue owners:
    # XXX Diverted requesters are not accorded full requester status here.
    reqid = issue.requester
    if reqid == 'Anonymous User':
        reqobj = acl_users._nobody
    else:
        reqobj = trackerGetUser(item, reqid)

    if issue.alt_full_names and issue.alt_full_names.has_key(reqid):
        reqnm = "%s (%s)" % (issue.alt_full_names[reqid], reqid)
    elif hasattr(reqobj, 'full_name') and reqobj.full_name:
        reqnm = reqobj.full_name
    else:
        reqnm = reqid
    subsids = issue.subscribers
    issueowners = issueOwners(item)

    # Identify the message type:
    if item.msgrole == "correspond":
        msgrole = ''
    else:
        msgrole = ' ' + item.msgrole
    if not hasattr(tracker, 'abbrev'):
        tracker.abbrev = ''
    baseemblem = "%s(%s)[%s]" % (tracker.abbrev, issue.id, item.id)
    plainemblem = "[%s]" % baseemblem
    roleemblem = "[%s%s]" % (baseemblem, msgrole)

    # Identify the role(s) of the user sending the message :
    isrequester = isstaff = isowner = 0
    if tracker_role(item, 'staff', user=authuser):
        isstaff = 1
    if tracker_role(item, 'Requester', user=authuser):
        isrequester = 1
    if tracker_role(item, 'IssueOwner', user=authuser):
        isowner = 1

    # Produce the message origin (from) and recipients (to, cc) entries.
    # (The actual destinations are passed to the smtp server separately, these
    # must direct replies to the right place - some person or program that
    # will say they need to use the web interface to reply.)
    reqbanner = "Issue Requester"
    reqccbanner = "Requester Courtesy CC"
    tobanner = frombanner = ccbanner = ''
    toids, ccids = [], []
    noisy = (hasattr(tracker, 'noisy') and tracker.noisy)
    if noisy or not issueowners or wasstage == 'pending':
        # Send to relevant supporters when noisy, unowned, or leaving pending.
        relevants = relevant_supporters(item)
        if not relevants:
            relevants = [tracker.adminaddr]
        if isrequester:
            frombanner = "%s - %s" % (usernm, reqbanner)
            tobanner = "Candidate Supporters"
            toids = relevants
            ccids = [userid]
            ccbanner = reqccbanner
        elif isstaff:
            frombanner = "%s - Tracker Staff" % usernm
            if item.private:
                tobanner = "Private item - non-staff excluded"
            else:
                tobanner = reqbanner
                toids = [issue.requester]
            ccids = relevants
            ccbanner = "Candidate Supporters"
    else:
        if isrequester:
            frombanner = "%s - %s" % (usernm, reqbanner)
            tobanner = "%s - Issue Supporter" % usernm
            toids = issueowners
            ccids = [userid]
            ccbanner = reqccbanner
        elif isstaff:
            frombanner = "%s - Issue Supporter" % usernm
            if item.private:
                tobanner = "Private item - non-staff excluded"
            else:
                tobanner = reqbanner
                toids = [issue.requester]
            ccids = issueowners
            ccbanner = "Supporter Courtesy CC"
    if not item.private and subsids:
        ccids = ccids + subsids

    # Determine recipient addresses:
    # (Some elaboration to cull duplicates and members lacking email fields.)
    # Resolve the total collection of recipients
    totalto = []
    badids = []
    for i in unique(toids + ccids):
       gotemail = ''
       if issue.alt_emails and issue.alt_emails.has_key(i):
           totalto.append(issue.alt_emails[i])
           continue
       memberobj = trackerGetUser(item, i)
       if memberobj and hasattr(memberobj, 'email') and memberobj.email:
           totalto.append(memberobj.email)
       else:
           badids.append(i)
    # Note bad ids, for admin info:
    def badsubs(ids, badids):
        newids = []
        for i in ids:
            if i in badids: newids.append('_' + i + '_')
            else: newids.append(i)
        return newids
    if badids:
        toids = badsubs(toids, badids)
        ccids = badsubs(ccids, badids)

    # Compose the message:

    d = DateTime()
    text = (('Subject: %s %s\n' % (roleemblem, item.title))
            + 'Sender: <%s>\n' % sender
            + 'Errors-To: <%s>\n' % sender
            + ('From: "%s, by Tracker - admin:" <%s>\n'
               % (singlifyQuotes(frombanner), sender))
            + ('To: "%s, via Tracker %s" <>\n'
               % (singlifyQuotes(tobanner), tracker.absolute_url()))
            + ('Cc: "%s, via Tracker %s" <>\n'
               % (singlifyQuotes(ccbanner), tracker.absolute_url()))
            + 'Date: %s\n' % d.rfc822()
            )
    if badids:
        text = (text
                + "X-Tracker-errors: Email-less members: %s\n" % badids)
    if RELEVANCE_DEBUG:
        text = text + "X-Tracker-debug: To: %s, Cc: %s\n" % (toids, ccids)
##        text = text + ("X-Tracker-debug: To: %s, Cc: %s; %s\n"
##                       % (toids, ccids, totalto))

    text = text + "\n"                  # Headers/body break.

    text = text + "Tracker Item %s - " % baseemblem
    diverted_to = None
    if issue.stage == 'diverted':
        diverted_to = hrefize(issue, issue.diverted, forcelink=1,
                              REQUEST=REQUEST)[1][0]
        div_tracker, div_issue, div_item = get_elements(diverted_to)
        divemblem = "%s(%s)[]" % (div_tracker.abbrev, div_issue.id)
        tense = ((msgrole == 'diverted') and "is being") or "has been"
        text = text + "The issue %s subsumed in %s\n" % (tense, divemblem)
        text = text + "which is at:\n    %s\n" % div_issue.absolute_url()
        text = text + ("The diverted item is at:\n    %s\n"
                       % item.absolute_url())
    else:
        text = text + "to followup, visit\n    %s\n" % item.absolute_url()
    references = item.references
    # Incidentals - references and attachments:
    if issue.stage == 'diverted': non_divert_refs = references[1:]
    else: non_divert_refs = references
    if non_divert_refs and non_divert_refs[0]:
        text = text + "References: "
        for i in non_divert_refs:
            objs = hrefize(issue, i, emblem_url_only=1, REQUEST=REQUEST)[1]
            if objs:
                i = ''
                for o in objs:
                    i = ((i and i + ' ') or '') + o.absolute_url()
            text = text + '\n    ' + i
        text = text + '\n'
    attachments = item.objectItems()
    if attachments:
        text = text + "Attachments:"
        for i in attachments:
            url = i[1].absolute_url()
            text = text + "\n    " + url
        text = text + '\n'


    # Body:
    text = text + ("\t\t-=-\t\t-=-\t\t-=-\n"
                   + "%s\n" % item.description
                   + "\t\t-=-\t\t-=-\t\t-=-\n")

    # Request status: Requester/company, stage, supporter
    text = text + 'Issue Characteristics:\n\n'
    text = text + 'Title: %s\n' % issue.title
    status = "Requester: %s \nStage: %s" % (reqid, capwords(issue.stage))
    if wasstage != issue.stage:
        status = status + " (from %s)" % wasstage
    if issue.stage != 'pending' and issueowners:
        status = status + ",  Supporter:"
        for i in issueowners:
            status = status + " " + i
    text = text + status + "\n"

    # Traits:
    traits = "Traits:\n        "
    first = 1
    for t, v in issue.traitVals:
        if not first: traits = traits + ", "
        else: first = 0
        addition = capwords(t) + ": " + v
        if len(traits + addition) > 70:
            text = text + traits + "\n"
            traits = "        "
        traits = traits + addition
    text = text + traits + "\n"

    totalto = unique(totalto)
    __traceback_info__ = ('to', totalto, 'toids', toids, 'ccids', ccids)
    if not totalto:
        totalto = [tracker.adminaddr]
        text = ("NOTE - this message was sent to the tracker administrator\n"
                "       due to lack of any valid addresses among the proper\n"
                "       recipients:\n\n"
                "       Member recipients lacking email addr: %(badids)s\n"
                "       Please visit the tracker site and rectify this:\n"
                "    %(adminurl)s\n\n" % {'badids': badids,
                                          'adminurl': tracker.absolute_url()}
                + text)

    send_email(tracker, sender, totalto, text)

    # Register the transmission in the audit trail:
    role = ''
    if isrequester:
        role = 'requester'
    if isstaff:
        if isowner:
            staffrole = 'owner'
        else: staffrole = 'staff'
        if role:
            role = role + '/' + staffrole
        else: role = staffrole
    issue.audited_state.set('msgfromrole', role, userid)
    # Phew.  Done.

def specialItemKnobs(self, REQUEST=None):
    """Promote or demote an item."""
    if REQUEST is None:
        return
    if REQUEST.get('new', None) == "Create a New Message":
        REQUEST.set('newitem', 1)
        return REQUEST['RESPONSE'].redirect(self.absolute_url() + "?newitem=1")
    requester = `REQUEST.AUTHENTICATED_USER`
    msgrole = REQUEST.get('msgrole', 'correspond')

    # Privacy stuff:
    privacy = REQUEST.get('privacy', 0)
    if privacy:
        if (privacy == "private") and not self.private:
            if int(self.id) in self.promotes():
                raise 'Bad Request', "Cannot conceal summary items."
            set_privacy(self, 1, asissue=0)
        elif (privacy == "public") and self.private:
            set_privacy(self, 0, asissue=0)

    if REQUEST.get('privacy_only', None) is not None:
        return REQUEST['RESPONSE'].redirect(self.absolute_url())

    direction = REQUEST.get('promoting', None)
    if direction is not None:
        if direction == "Promote":
            if self.private:
                raise 'Bad Request', "Summary items cannot be private."
            self.msgrole = msgrole
            # Adjust item privacy congruent with issue as a whole:
            set_privacy(self, self.aq_parent.private, asissue=1)
            # ... and it cannot be staff-private if it's part of the summary:
            self.private = 0
            self.promotes(self.id, requester)
            if msgrole == 'reframe':
                issue = self.aq_parent
                issue.reframed = self.id
                issue.description = self.description
                issue.title = self.title
        elif direction == "Demote":
            self.msgrole = msgrole
            self.promotes(self.id, requester, demote=1)
            issue = self.aq_parent
            if issue.reframed == id:
                request = issue.get(1)
                issue.description = request.description
                issue.title = request.title
        else:
            raise 'Bad Request', ("Unrecognized item promotion, "
                                  + `direction`)

    return REQUEST['RESPONSE'].redirect(self.absolute_url())

def setIssueStage(self, stage, REQUEST, no_promote=0):
    """Do issue stage-transition business, from items and issues.

    Optional no_promote means don't promote the item to the summary, even for
    summary-worthy actions.

    XXX This code constrains issue stage *workflow* transitions.  However, the
    actual choices offered on the message forms are controlled by dtml in the
    itemForm.  The workflow transition logic should be centralized."""

    if stage not in self.stage_range:
        raise 'Bad Request', "Unrecognized item stage %s" % `stage`
    action = REQUEST.get('msgrole', '')
    user = (REQUEST and REQUEST.AUTHENTICATED_USER)
    userid = (user and `user`) or ''
    tracker, issue, item = get_elements(self)
    if not issue:
        raise 'Bad Request', "No containing issue found."

    # Verify authorization for stage change:
    if issue.stage != stage:
        which = (action and "the %s" % action) or "this"
        if userid not in (issueOwners(issue) + tracker.supporters):
            if stage == 'rejected':
                # Ok for issue requester (original or deferred) and supporters.
                if userid == issue.requester:
                    if issue.stage != 'pending':
                        raise 'Bad Request', ("Requesters may not reject"
                                              " issues once they're no longer"
                                              " pending.")
                elif not tracker_role(issue, 'staff', user=user):
                    raise 'Bad Request', ("Only issue requesters and"
                                          " tracker staff can take %s"
                                          " action." % which)
            else:
                raise 'Bad Request', ("You must be a supporter or an issue"
                                      "\nowner to take %s action." % which)

    if stage in ['accepted', 'resolved', 'completed', 'diverted', 'deferred',
                 'rejected']: 
        if userid in tracker.supporters:
            issue.manage_addLocalRoles(userid, ['IssueOwner'])
        if item:
            item.msgrole = stage
    elif stage == 'pending':
        # Enable moving an issue to pending while still keeping
        # supporter assignment.  People can 'Edit State' to remove the 
        # supporter, as well.
        # XXX Eventually this should be an optional behavior.
        #clear_local_role('IssueOwner', issue)
        if item:
            msgrole = item.msgrole
            if item.msgrole == 'accept':
                item.msgrole = 'correspond'
    if (item is not None) and not no_promote:
        if item.private:
                raise 'Bad Request', ("Action summary items cannot be private."
                                      '\n - either select "Don\'t post" or'
                                      " submit a non-action message.")
        issue.promotes(item.id, userid)

    REQUEST.set('stage', stage)
    reconcileState(issue, REQUEST)

def snarfAttachments(REQUEST):
    attachments = []
    errmsg = ""
    for i in range(int(REQUEST.get('slots', 1))):
        t = REQUEST.get('file%stitle' % i, None)
        f = REQUEST.get('file%s' % i, None)
        if t or (f and f.filename):
            if t and not (f and f.filename):
                if errmsg: errmsg = errmsg + "; "
                errmsg = errmsg + ("Issue submitted with file title '%s'"
                                   " but no file." % t)
            id, title = OFS.Image.cookId(None, t, f)
            attachments.append((id, t, title, f))
    return attachments, errmsg

def reconcileState(self, REQUEST=None):
    """Register setttings for audited attrs, triggering audit and rules."""
    tracker, issue, item = get_elements(self)
    setTraitValues(issue, tracker, REQUEST=REQUEST)
    if REQUEST is not None:
        # Attribute set:
        #  {'setName1': ([attr1, ...], {attr1: [val1, ...], ...}),
        #   'setName2': ...}
        # E.g.:
        #  {'StatusVars': (['stage', 'priority', ...],
        #                  {'stage', ['pending', 'accepted', ...]},
        #                  {'priority', ['normal', 'low', 'high', ...]},
        #                  {...})}
        for attrSet in issue.auditedAttrs.values():
            for attrName in attrSet[0]:
                # Yikes!  We don't do the range checking for which these vars
                # were established.
                if (hasattr(REQUEST, attrName)
                    and (getattr(issue, attrName) != REQUEST.get(attrName))):
                    setattr(issue, attrName, REQUEST.get(attrName))
    statechanged = issue.register_state_changes(REQUEST)

    if REQUEST.get('editState', REQUEST.get('issueIndex', None)) is not None:
        # Reconcile supporter, requester, traits settings:
        verified = []
        new = REQUEST.get('open_subscribe', None)
        if (new is not None and hasattr(issue, 'open_subscribe')):
            if type(new) == type(''):
                if string.lower(new) == "on":
                    new = 1
                elif string.lower(new) == "off":
                    new = 0
                else:
                    __traceback_info__ = new
                    new = int(new)
            issue.open_subscribe = new
        new = REQUEST.get('subscribers_may_post', None)
        if (new is not None and hasattr(issue, 'subscribers_may_post')):
            if type(new) == type(''):
                new = int(new)
            issue.subscribers_may_post = new
        new = REQUEST.get('requester', None)
        if ((new is not None) and hasattr(issue, 'requester')
            and new != issue.requester):
            if not verifyUser(issue, new, verified):
                raise "Invalid User", "Invalid user '%s'" % new
            clear_local_role('Requester', issue, user=issue.requester)
            issue.manage_addLocalRoles(new, ['Requester'])
            issue.requester = new
        new = REQUEST.get('owners', None)
        if ((new is not None) and (new and hasattr(issue, 'issueOwners'))):
            old = issue.issueOwners()
            new.sort(); old.sort()
            if new != old:
                # Note - we allow users not registered as tracker supporters
                # to get the role - handy for per-issue supporting...
                for i in new:
                    if not verifyUser(issue, i, verified):
                        raise "Invalid Id", ("Invalid id '%s' specified"
                                             " for supporters" % i)
                added, defunct = diffs(new, old)
                for i in defunct:
                    clear_local_role('IssueOwner', issue, user=i)
                for i in added:
                    issue.manage_addLocalRoles(i, ['IssueOwner'])
        new = REQUEST.get('subscribers', None)
        if ((new is not None) and hasattr(issue, 'subscribers')):
            new = filter(string.strip, new)
            old = issue.subscribers[:]
            new.sort(); old.sort()
            if new != old:
                for i in new:
                    if not verifyUser(issue, i, verified):
                        raise "Invalid Id", ("Invalid id '%s' specified"
                                             " for subscribers" % i)
                added, defunct = diffs(new, old)
                for i in defunct:
                    clear_local_role('Subscriber', issue, user=i)
                for i in added:
                    issue.manage_addLocalRoles(i, ['Subscriber'])
                issue.subscribers = new
        new = REQUEST.get('subscribe', None)
        if ((new is not None) and hasattr(issue, 'subscribers')):
            user = REQUEST.AUTHENTICATED_USER
            issubscriber = tracker_role(issue, 'Subscriber', user=user)
            subscribers = issue.subscribers
            if new and not issubscriber:
                issue.manage_addLocalRoles(`user`, ['Subscriber'])
                if `user` not in subscribers:
                    subscribers.append(`user`)
                    issue.subscribers = subscribers
            elif not new and issubscriber:
                clear_local_role('Subscriber', issue, user=`user`)
                if `user` in subscribers:
                    subscribers.remove(`user`)
                    issue.subscribers = subscribers

##    if REQUEST.get('subscribe', None):
##        issue.manage_addLocalRoles(`REQUEST.AUTHENTICATED_USER`,
##                                   ['Subscriber'])
##        old = issue.subscribers
##        issue.subscribers.append(`REQUEST.AUTHENTICATED_USER`)
##        issue.subscribers = issue.subscribers
    return statechanged

def set_privacy(self, restrict, asissue=0):
    """Activate or deactivate element privacy as for an issue or item.

    The eligible public for elements without privacy restrictions depends on
    the tracker type.  In an open tracker, the public includes site members
    and the anonymous user, as well as tracker staff, etc.  In a dedicated
    tracker, the tracker staff, the requester, and designated tracker clients
    consititue the public - others are excluded even from private elements.

    Privacy at the issue level means restricted to requester and observers
    (which includes staff and subscribers).

    Privacy at the item level means restricted to staff, excluding
    requester and subscribers.

    When promoting an item as part of the issue summary, its privacy should be
    set as for an issue, according to the privacy setting of the issue."""

    self.private = restrict
    minimal = ROLE_CATEGORIES['staff']
    intimates = unique(ROLE_CATEGORIES['observer'] + minimal)
    maximal = ROLE_CATEGORIES[self.access_mode or 'open']

    if not restrict:
        if asissue:
            self.manage_permission('View', roles=maximal, acquire=0)
        else:
            self.manage_permission('View', roles=intimates, acquire=0)
    else:
        if asissue:
            self.manage_permission('View', roles=intimates, acquire=0)
        else:
            if self.id in self.promotes():
                raise 'Bad Request', "Promoted items must be public."
            self.manage_permission('View', roles=minimal, acquire=0)
        

## def index_object(self):
##     """Site catalog indexing

##     The tracker has an explicit one for the reindexing process, so that the 
##     subcomponents are traversed then."""
##     return None                         # XXX !!! Temporary until cataloging

##     __traceback_info__ = (`self`,)
##     for i in self.values():
##         for j in i.values():
##             if hasattr(j, '_doindex') and j._doindex:
##                 j.unindex_object()
##                 j.index_object()
##         i.unindex_object()
##         i.index_object()
##     self.SiteIndex.indexObject(self)

def trackerOnDeleteObject(self):
    """Object delete handler.

    Tracker has an explicit one both to do its own unindexing and also cause 
    traversal of the subcomponents."""
    for i in self.values():
        for j in i.values():
            j.unindex_object()
        i.unindex_object()

# Catalog/searching stuff ================================================

def trackerInitCatalog(self, REQUEST=None):
    self._catalog = Catalog()
    self.threshold = 1000
    self._v_threshold = 0

    self._catalog.addColumn('type')
    self._catalog.addIndex('type', 'FieldIndex')
    ## XXX Eventually: self._catalog.manage_addColumn('type')

    self._catalog.addIndex('numid', 'FieldIndex')

    self._catalog.addIndex('exemptions', 'FieldIndex')

    self._catalog.addIndex('text_content', 'TextIndex')

    self._catalog.addIndex('date', 'FieldIndex')

    self._catalog.addIndex('stage', 'FieldIndex')

    self._catalog.addIndex('priority', 'FieldIndex')

    self._catalog.addIndex('private', 'FieldIndex')

    self._catalog.addIndex('issueOwners', 'KeywordIndex')

    self._catalog.addIndex('requester', 'FieldIndex')

    self._catalog.addIndex('subscribers', 'KeywordIndex')

    self._catalog.addIndex('from', 'FieldIndex')

    # XXX Eventually this will need to be a separate index for each trait.
    #     Problem is that the collection of traits is dynamic, and the names
    #     may well collide with catalogued attributes, like title, priority,
    #     stage, etc.  So we would need a handle in each element for each
    #     trait - yuck.  For now they're grouped together - which means people 
    #     only get an 'or' style search within traits.  If we had some easy
    #     'and'ing mechanism for search results, we would have a way to do
    #     multiple passes, refiningthe search to the desired results.
    self._catalog.addIndex('traitVals', 'KeywordIndex')

    tracker_path = self.absolute_url()
    for issue in self.values():
        catalogElement(issue, recursive=1)

def element_text_content(self):
    """Return the text content of the object, for consumption by catalog."""
    content = self.title + ' ' + self.description
    if self.type == 'tracker_issue':
        # Should do a list accumulation then a string join.
        for i in self.values():
            content = content + ' ' + element_text_content(i)
    return content

def catalogElement(element, recursive=0):
    "Catalog an element, and its items if recursive true and it's an issue."
    tracker, issue, item = get_elements(element)
    if element.type == 'tracker_issue':
        tracker._catalog.catalog_object(issue, make_catalogPath(issue))
        if recursive:
            for item in issue.values():
                tracker._catalog.catalogObject(item, make_catalogPath(item))
    else:
        tracker._catalog.catalogObject(item, make_catalogPath(item))

def recatalogElement(self, REQUEST=None):
    """Refresh the cataloging of issue or element."""
    element = self
    tracker, issue, item = get_elements(element)
    path = make_catalogPath(element)
    tracker._catalog.uncatalogObject(path)
    tracker._catalog.catalogObject(element, path)
    
def make_catalogPath(element):
    """Construct element's catalog path, wrt encompassing tracker and issue."""
    if element.type == 'tracker_issue':
        return str(element.id)
    elif element.type == 'tracker_item':
        return str(element.aq_parent.id) + '/' + str(element.id)

def catalogPath_to_object(self, search_result):
    """Get element corresponding to catalog path."""
    path = self._catalog.paths[search_result.data_record_id_]
    splat = string.split(path, '/')
    issue = self[splat[0]]
    if len(splat) == 2:
        return issue[splat[1]]
    else:
        return issue

def issuesSearch(self, features, REQUEST=None, viewable_only=0,
                 split=string.split, strip=string.strip):
    """Find issues according to feature sets.

    Returns a triple with the issue batch, the number of preceeding, and the
    number of succeeding issues.

    Multiple choices within a field are 'or'ed together, no choices for a
    field means that field is disregarded for the search, and the hits across
    a collection of specified fields is 'and'ed together.

    Special feature 'any' is targeted to the text_content attribute of the
    elements - typically a method that sums the title and description field.

    Supported fields:

      any: text query on text_content
      type: 'tracker_issue', 'tracker_item'
      requester: fields index
      private, public:  'ON' or ()
      supporter: list of supporters: [supporter, ...]
      stage: which stage(s) - []
      priority: []
      refine_traits: {trait: [vals], ...} select traits with attr values.
      since: datestr - elements subsequent to specified date.
      until: datestr - elements prior to specified date.
      batch_size: max number of issues to return
      jump: target item number for jump, or 0
      batch_direction: -1, 0, 1 for backwards, stay, or forwards
      batch_last_start: start id of the prior batch
      batch_last_end: end id of the prior page
      """

    tracker, issue, item = get_elements(self)
    catalog = tracker._catalog

    traits = features.get('refine_traits', {})
    traitVals = []
    for t, v in traits.items():
        # Populate with [(trait, val), (trait, val2), ...]
        if v:
            for i in v:
                traitVals.append((t, i))
    criteria = {
        #'sort_on': 'numid',	# XXX ZCatalog bug in combo w/TextIndex
        'type': features.get('type', []),
        'stage': features.get('stage', []),
        'issueOwners': features.get('supporter', []),
        'requester': features.get('requester', ''),
        'traitVals': traitVals,
        'priority': features.get('priority', []),
        }

    anys = features.get('any', '') or ''
    if anys:
        criteria['text_content'] = anys

    private = features.get('private', 1) or 0
    public = features.get('public', 1) or 0
    if (public or private) and not (public and private):
        if private:
            criteria['private'] = [1, 'on']
        else:
            criteria['private'] = [0, 'off']

    until_active = features.get('until_active', '')
    since_active = features.get('since_active', '')
    if until_active or since_active:
        datespan = []
        criteria['date'] = datespan
        if since_active:
            sincedate = DateTime("%s %s, %s" %
                                 (REQUEST.get('since_month'),
                                  REQUEST.get('since_day'),
                                  REQUEST.get('since_year'))).earliestTime()
            datespan.append(sincedate)
            REQUEST.set('since_date', str(sincedate))
        if until_active:
            untildate = DateTime("%s %s, %s" %
                                 (REQUEST.get('until_month'),
                                  REQUEST.get('until_day'),
                                  REQUEST.get('until_year'))).latestTime()
            datespan.append(untildate)
            REQUEST.set('until_date', str(untildate))
        if until_active and since_active:
            criteria['date_usage'] = 'range:min:max'
        elif until_active:
            criteria['date_usage'] = 'range:max'
        else:
            criteria['date_usage'] = 'range:min'

    for k, v in criteria.items():
        if not v: del criteria[k]

    got_ids = []
    cpaths = catalog.paths

    results = catalog.searchResults(criteria)
    for i in results:
        got_ids.append(cpaths[i.data_record_id_])
    # XXX In lieu of broken catalog sort_on (with TextIndexes), sigh:
    if 1:
        def sort_ids(x, y, atoi=string.atoi):
            x, y = atoi(x), atoi(y)
            return (((x < y) and -1) or ((x == y) and 0) or 1)
        got_ids.sort(sort_ids)

    preceeding = succeeding = 0
    batch_size = features.get('batch_size', 0)
    if batch_size:
        culled = batchcull(got_ids,
                           size=batch_size,
                           jump=features.get('jump',0),
                           direction=features.get('batch_direction', 0),
                           last_start=features.get('batch_last_start', 0),
                           last_end=features.get('batch_last_end',
                                                 len(got_ids)))
        got_ids, preceeding, succeeding = culled
    got = []
##    # DEBUG
##    if REQUEST.AUTHENTICATED_USER.name == 'klmtest':
##        import pdb
##        pdb.set_trace()
    for i in got_ids:
        anissue = tracker[i]
        if viewable_only and not viewable(anissue, REQUEST):
            continue
        got.append(anissue)

    return (got, preceeding, succeeding)

def batchcull(pool,
              size=0,
              jump=0,
              direction=0,
              last_start=0, last_end=-1):
    """Return before/after/at prev batch, and figures for remainder:

    triple: ([batch...], num-preceeding, num-succeeding)

    pool is the entire pool of ids to select among.

    size is the batch size - 0 means all.

    direction indicates:

     -1: prior to start of current batch
     0: at current batch (but possibly different number)
     1: subsequent of current batch.

    last_start and last_end indicate boundaries of previous batch.  A value
    of '-1' means the last element."""

    __traceback_info__ = (size, direction, last_start, last_end)
    if type(size) == type(""): size = string.atoi(size)
    if type(jump) == type(""): jump = string.atoi(jump)
    if type(direction) == type(""): direction = string.atoi(direction)
    if type(last_start) == type(""): last_start = string.atoi(last_start)
    if type(last_end) == type(""): last_end = string.atoi(last_end)

    if not size or not pool: return (pool, 0, 0)
    lenpool = len(pool)

    if direction in [-1, 0]:
        boundary = last_start
    else:
        boundary = min(last_end, len(pool))
    if boundary == -1:
        boundary = max(0, len(pool) - 1)
    if boundary > lenpool:
        boundary == lenpool
    
    if direction in [0, 1]:
        # Going forwards from boundary.
        start = boundary
        end = boundary + size
    else:
        # Going backwards from boundary.
        end = max(min(boundary, lenpool), 0)
        start = max(end - size, 0)
        end = start + size

    if end > lenpool: end = lenpool

##    raise "DEBUG", ("p:%s, bs:%s, d:%s, ls:%s, le:%s, b:%s, st:%s, e:%s"
##                    % (len(pool), size, direction, last_start, last_end,
##                       boundary, start, end))
    return(pool[start:end], start, len(pool) - end)
    
def fieldValuesRange(self, field, precedence=None, include=None):
    """Return the sorted range of values for a particular catalog field.

    Optional arg 'precedence' is a list of fields that are to come first in the
    sorted results, if present.

    Optional arg 'include' is a list of fields from the precedence elements to
    be included even if not present among the existing catalog values.

    Useful for presenting subsets of a range (such as the values within a
    range that occur in a catalog field index) in a particular ordering."""

    tracker, issue, item = get_elements(self)
    try:
        uniquevals = tracker._catalog.uniqueValuesFor(field)
    except KeyError:
        return []
    if not precedence:
        got = list(uniquevals)
        got.sort()
        return got
    else:
        got = {}
        __traceback_info__ = (field, uniquevals)
        for i in uniquevals: got[i] = 1
        ordered = []
        hasit = got.has_key
        for i in precedence:
            if hasit(i):
                ordered.append(i)
                del got[i]
            elif include and (i in include):
                ordered.append(i)
        if got:
            got = got.keys()
            got.sort()
            return ordered + got
        else:
            return ordered

# Preferences stuff ================================================

# I use this funky NOVAL for default values because sometimes a None needs
# to be passed in, so it won't do.  I may be forgetting a smarter way.
NOVAL = '<Unset doggonit!>'
def trackerpref(user=None, pref=NOVAL, value=NOVAL, default="", REQUEST=None):
    """Get or set a user preference for application app.

    Get if no value is specified, set otherwise.

    Value of 'default' arg (empty string if not passed in) is returned for
    gets of unset vars."""

    # XXX Would probably be better to use an explicit .get/.set
    #     interface, if this were actually *on* a user object.
    # XXX Generalizing this would means something like 'MemberPref', where the 
    #     application name would be a parameter - and there should be some
    #     shared space.
    # XXX There's no security provision!  Not sure how i'd organize it.  Sigh.

    prefs = 'preferences'
    app = 'tracker'
    if user is None:
        if REQUEST is None:
            raise ValueError, "No REQUEST passed in."
        user = REQUEST.AUTHENTICATED_USER

    acl_users = user.acl_users

    if pref == NOVAL:
        raise ValueError, "No preference specified."

    if not hasattr(user, prefs):
        prefsdict = {}
    else:
        prefsdict = getattr(user, prefs)
    if not prefsdict.has_key(app):
        # Initiate app dictionary for app if not already there.
        appdict = {}
        prefsdict[app] = appdict
    else:
        appdict = prefsdict[app]
    if value == NOVAL:
        if appdict.has_key(pref):
            return appdict[pref]
        else:
            return default
    elif user in [acl_users._nobody, acl_users._super]:
        # Setting, but to a builtin user:
        raise ValueError, "User prefs not allowed for anonymous or superuser."
    else:
        # Setting, and to a regular user:
        if not prefsdict.has_key(app):
            prefsdict[app] = appdict
        appdict[pref] = value
        user.preferences = prefsdict

# Report stats stuff ================================================

def quota_stat_search(tracker, criteria=None, ignore_date=0, expired=(),
                      period_only=None):
    """Get the search results satisfying a tracker quota stat.

    We return those within the period of service (for dedicated trackers) or
    over the course of the preceeding year (for open trackers), unless
    ignore_date is set, in which case an unlimited span is respected.

    The following assumes that ignore_date is unset.

    If a dedicated tracker and outside the service period, then 'expired's
    value (default: ()) is returned.

    If period_only is set, return just a tuple of the dates - (start, end)
    that would be used to constrain the results for the non-period_only
    version of the query."""

    if ignore_date:
        if period_only:
            return (0, 0)
        return tracker._catalog.searchResults(criteria)

    dedicated = (tracker.access_mode == 'dedicated')

    now = DateTime()

    if dedicated:
        start = tracker.tickets_start_date.earliestTime()
        end = tracker.tickets_end_date.latestTime()
        if period_only:
            return (start, end)
        if (start > now.earliestTime()) or (end < now.latestTime()):
            return expired
    else:
        start = (now - 365).earliestTime()
        end = now.latestTime()
        if period_only:
            return (start, end)

    
    criteria = criteria.copy()
    criteria.update({'date': [start, end], 'date_usage': 'range:min:max'})

    results = tracker._catalog.searchResults(criteria)
    return (results, start, end)

INFINITY = 1e200**2

def quota_stats(self, stat, expired=0):
    """Return triple: (tracker activity stat, start-date, end-date).

    An activity stat is the number of tracker issues having the indicated
    status - for 

    The figures are computed with respect to the tracker's quota, if it's a
    dedicated one.

    The stat is one of:

     allotted	    - Total number allotted
     remaining	    - Number of available tickets remaining
     period         - Return the tracker's service period
     <stages>       - one of:

          pending, accepted, resolved, rejected, deferred, diverted. """

    tracker, issue, item = get_elements(self)
    sR = tracker._catalog.searchResults
    criteria = {'type': 'tracker_issue'}

    dedicated = (tracker.access_mode == 'dedicated')

    if stat == 'allotted':
        if not dedicated or tracker.tickets_max < 0:
            return sys.maxint
        else:
            now = DateTime()
            start = tracker.tickets_start_date
            end = tracker.tickets_end_date
            if (start > now.earliestTime()) or (end < now.latestTime()):
                return expired
            else:
                return tracker.tickets_max

    elif stat == 'remaining':
        resolved = quota_stats(tracker, 'resolved', expired=-1)
        if resolved == -1:
            # If we're outside the service dates, we clamp down.
            return expired
        if not dedicated or tracker.tickets_max < 0:
            return sys.maxint
        else:
            return (tracker.tickets_max -
                    (resolved - quota_stats(tracker, 'accepted')))

    # This needs to be after the 
    elif stat == 'period':
        return quota_stat_search(tracker, period_only=1)

    elif stat in tracker.stage_range:
        criteria.update({'stage': stat})
        result = quota_stat_search(tracker, criteria, expired=-1)
        if result == -1:
            return expired
        else:
            got, start, end = result
            return len(got)

    else:
        raise ValueError, "Unrecognized quota stat '%s'" % stat

def supporter_activity_stats(trackers):
    """Tally and return supporter ticket activity for a list of trackers.

    The activities are distinguished according to active (pending or
    accepted), resolved, and other (rejected, deferred, or
    diverted) tickets, versus support level.

    Two records are returned - at totals tally and a heavily nested
    per-supporter record:

    (<totals rec>, <supporters list>)
    <totals rec> := <activity rec>
    <supporters list>: ((supporter, <activity rec>), ...)
    <activity rec> := {'platinum': <tickets rec>, 'gold': <tickets rec>, 
                       'silver': <tickets rec>, 'bronze': <tickets rec>,
                       'open': <tickets rec>, 'total': <tickets rec>]
    <tickets rec> := [#active, #resolved, #other]."""

    # Glean the supporters across all trackers:
    supporters = {}
    totals = {'platinum': [0, 0, 0], 'gold': [0, 0, 0], 'silver': [0, 0, 0],
              'bronze': [0, 0, 0], 'open': [0, 0, 0], 'total': [0, 0, 0]}
    for t in trackers:
        catalog = t._catalog
        dedicated = t.access_mode == 'dedicated'
        if dedicated:
            start = t.tickets_start_date.earliestTime()
            end = t.tickets_end_date.latestTime()
        else:
            now = DateTime()
            start = (now - 365).earliestTime()
            end = now.latestTime()
        criteria = {'date': [start, end], 'date_usage': 'range:min:max',
                    'type': 'tracker_issue'}
        for s in t.supporters:
            if not supporters.has_key(s):
                supporters[s] = {'platinum': [0, 0, 0], 'gold': [0, 0, 0],
                                 'silver': [0, 0, 0], 'bronze': [0, 0, 0],
                                 'open': [0, 0, 0], 'total': [0, 0, 0]}
            level = (dedicated and t.support_level) or 'open'
            rec = supporters[s][level]
            levtot = totals[level]
            tottot = totals['total']
            suptot = supporters[s]['total']
            criteria['issueOwners'] = s

            for i, stages in ((0, ['pending', 'accepted']),
                              (1, ['resolved']),
                              (2, ['rejected', 'deferred', 'diverted'])):
                criteria['stage'] = stages
                add = len(catalog.searchResults(criteria))
                rec[i] = rec[i] + add
                levtot[i] = levtot[i] + add
                suptot[i] = suptot[i] + add
                tottot[i] = tottot[i] + add
    slist = supporters.items()
    slist.sort()
    return slist + [('<b>Total</b>', totals)]

DAY = 1
WEEK = DAY * 7
SPANS = (("1 Day", DAY), ("1..2 Days", 2*DAY), ("..4 Days", 4*DAY),
         ("..1 Week", 1*WEEK), ("..2 Weeks", 2*WEEK), ("..3 Weeks", 3*WEEK),
         ("..4 Weeks", 4*WEEK), ("..8 Weeks", 8*WEEK),
         ("8+ Weeks", sys.maxint))

def issue_maturity_stats(trackers, stages):
    """Tally and return how many days tickets occupy their current stage.

    Trackers are the tracker objects to serve as the collective subject of the
    tally.  (The value can be an individual tracker instead of a list of
    them.)  Stages is a list of the stages of interest.  These will be the
    rows of the return value.  The columns are the age of the items in those
    stages.

    We return a list of lists:
    [["Stage", "2d", "4d", "1w", "2w", "3w", "4w", "8+w"],
     ["Pending", 3, 2, 0, 3, 0, ...],
     ["Accepted", 2, 1, 0, 4, 0, , ...], ...]."""

    if type(trackers) != type([]):
        trackers = [trackers]

    got = []
    header = ['<em>Stage</em>']
    totaldict = {}
    for legend, period in SPANS:
        header.append(legend)
        totaldict[period] = 0
    got.append(header)
    for s in stages:
        now = DateTime()
        issues = []
        criteria = {'type': 'tracker_issue', 'stage': s}
        for t in trackers:
            c = t._catalog
            issues = (issues
                      + map(lambda x, t=t, c=c: t[c.paths[x.data_record_id_]],
                            t._catalog.searchResults(criteria)))
        ages = []
        for i in issues:
            ages.append(now - i.date)
        ages.sort()                     # youngest to oldest
        tally = [string.capitalize(s)]
        for legend, period in SPANS[:-1]:
            periodtotal = 0
            while ages and (ages[0] <= period):
                periodtotal = periodtotal + 1
                del ages[0]
            tally.append(periodtotal)
            totaldict[period] = totaldict[period] + periodtotal
        # Residue:
        totaldict[SPANS[-1][1]] = totaldict[SPANS[-1][1]] + len(ages)
        tally.append(len(ages))

        got.append(tally)
    total = ["Total"]
    for legend, period in SPANS:
        total.append(totaldict[period])
    got.append(total)
    return got

def tracker_peers(self, peers=None, REQUEST=None, split=string.split):
    """Fish out peer tracker objects by their (relative) URLs.

    We return two elements in a tuple:

    0: A list of peer tracker objects
    1: (URL (object,)) tuples of indicated peers, adding new ones.

    Only tracker staff members are authorized to use this routine.

    peers should be a list of strings starting with URLs of peer trackers
    trackers to be considered.  Anything after whitespace following the URL is
    ignored (eg, the error notices described below can be included).

    The objects are returned as single-element tuples, to make them easy to
    distinguish from the error-message returns (ie, result[1][1:] == ()) and
    to insulate them from the zope auto-call mechanism applied in dtml-var and 
    dtml-in operations.

    For those URLs which do not resolve into trackers (or resolve into
    trackers in which the AUTHENTICATED_USER is not on staff), a string is
    returned in the object slot describing the problem:

      "<resolve_url exception type text>"
      "Not a tracker"
      "Tracker staff only"
      "Not on staff in this tracker"

    """

    tracker, issue, item = get_elements(self)

    if ((REQUEST is None)
        or
        (not tracker_role(tracker, 'staff',
                          user=REQUEST.AUTHENTICATED_USER))):
        raise "Unauthorized", "Tracker staff only."

    tracker, issue, item = get_elements(self)
    herepath = split(tracker.absolute_url(1), '/')  # Relative, not absolute.

    set_peers = filter(None, map(string.strip, peers))
    peers = []
    # Collect just the paths, eliminating duplicates in the process:
    for i in set_peers:
        ps = string.split(i)[0]
        if ps not in peers:
            peers.append(ps)

    got = []
    gotids = {}
    gotobjs = []                        # Use this to avoid repeats.
    for p in peers:
        if not p:
            continue
        if p and (p[0] == '/'): p = p[1:]

        ps = split(p, '/')

        #import pdb; pdb.set_trace()
        # Remove leading elements common with this tracker:
        for branchpoint in range(len(herepath)):
            if ((branchpoint >= len(ps))
                or (herepath[branchpoint] != ps[branchpoint])):
                break
        if branchpoint:
            ps = ps[branchpoint:]

        # Ascend to the branch point, then go back down to the object, bailing 
        # if we hit an access error.  *NOTE* that by ascending and descending
        # we ensure that the acqusition hierarchy - hence the permissions
        # context - is for the object we're visiting.
        msg = ''
        o = tracker
        try:
            for i in range(len(herepath) - branchpoint):
                o = o.aq_parent
            for i in ps:
                if i: o = getattr(o, i)
        except 'Unauthorized':
            o = None
            msg = 'No access to this object'
        except:
            o = None
            tbobj, tbmsg = sys.exc_info()[:2]
            # XXX Zope bug; shoulda been 'Unauthorized', above. klm 10/08/99
            if tbobj == 'Internal Server Error':
                msg = 'No access to this object'
            else:
                msg = str(tbobj)
            msg = msg + ', ' + str((herepath, branchpoint, ps))
            got.append(p, msg)

        if o is not None:
            if not hasattr(o, 'type') or o.type != 'tracker':
                got.append(o, "Not a tracker")
            else:
                if gotids.has_key(id(o.aq_base)):
                    # Already have one.
                    continue
                if o.tracker_role(o, 'staff', user=REQUEST.AUTHENTICATED_USER):
                    # Bingo - we have staff access in this object.
                    gotids[id(o.aq_base)] = 1
                    got.append(p, (o,))
                    gotobjs.append(o)
                else:
                    got.append(p, "Not staff in this tracker")
    tracker.peers = peers
    return (gotobjs, got)

# Membership Compat ==========================================
#
# In lieu of real membership, some routines that fit a prototype api.

def findAllMemberNames(self, ceiling=None):
    """Get all user names from all acl_users within scope.

    For now, every acl_user account is a member account.

    Optional ceiling specifies a maximum amount of users to fetch - we 
    return the usernames for the first of that number that we find."""
    # Collect users from all acquired acl_users:
    users = []
    for au in get_acl_users_acquired(self):
        users = users + au.user_names()
        if ceiling is not None and len(users) > ceiling:
            users = users[:abs(ceiling)]
            break
    # Get the distinct user names from the collection of users.
    got = {}
    for i in users:
        got[i] = None
    # Return the distinct user names.
    got = got.keys()
    got.sort()
    return got

def trackerGetUser(self, name, REQUEST=None):
    """acl_users.getUser() substitute, for dtml access to user objs."""
    # XXX This routine predates get_user_acquired(), or we'd use it directly.
    return get_user_acquired(self, name)

def editMembers(self, REQUEST=None):
    """Hack to make it easy for managers to set basic member attributes.

    XXX This is an expedient that gives improper power to the tracker
    admin to override users settings!  It should not be usable in a
    system with proper membership, where members can set their email
    and fullname settings.

    AUTHENTICATED_USER must have the 'manager' global role to run this.

    Go to the value of the REQUEST nextpage setting, if any, or to fromurl, if 
    any, or to referer."""

    if REQUEST is None:
        raise 'Bad Request', "No request passed in."

    requester = REQUEST.AUTHENTICATED_USER
    if not tracker_role(self, 'staff', user=requester):
        raise 'Bad Request', "Restricted to tracker owner and supporters." 

    for acl_users in get_acl_users_acquired(self):

        for membername in acl_users.user_names():
            member = acl_users.getUser(membername)
            for attr in ('full_name', 'email'):
                newval = REQUEST.get('_new_' + str(member) + '_' + attr, ())
                if newval is not ():
                    setattr(member, attr, newval)

    if REQUEST.has_key('nextpage'):
        return REQUEST['RESPONSE'].redirect(REQUEST.get('nextpage'))
    else:
        return REQUEST['RESPONSE'].redirect(
            string.split(REQUEST.get('REFERER', URL1), '?')[0]
            + '?manage_tabs_message=Changed!')

# Rules stuff ================================================

CONDITION = 'condition'
ACTIONS = 'actions'

STATE = 'state'
VALUE = 'value'
VOLATILE = 'volatile'
CONTAINS = 'contains'
TRANSITION = 'transition'
AND = 'and'
OR = 'or'
NOT = 'not'

SET = 'set'
INCLUDE = 'include'
REMOVE = 'remove'
NOTIFY = 'notify'
ABORT = 'abort'

ADAY = 1.0

def setTrackerRules(self, REQUEST=None):
    """Reset the tracker rules on the fly."""
    __traceback_info__ = (`self`, `self.set_rules`)
    phrasing = (self._rules is not None and "(re)set") or "set"
    self.init_expr_tables()
    rules = createRules()
    self.set_rules(rules)
    if REQUEST is not None:
        destmod = urllib.quote('<font color="green">Rules %s.</font>'
                               % phrasing)
        return REQUEST['RESPONSE'].redirect(REQUEST['URL1']
                                            + '?manage_tabs_message='
                                            + destmod)
    else:
        return rules

def createRules():
    """Create the rules for an issue tracker."""
    # XXX This is an external method for convenience.
    # XXX Eventually this will be a (perhaps more genericish) default, 
    #     and there will be a ttw interface for changing and refining them.

    # See TrackerBase.Noodge.eval_contingency(), .eval_condition(), and
    # .apply_actions() for contingency structure.
    standard = [
##        ( # Forbid transition from 'pending' to non-pending states.
##            (CONDITION, (AND, (STATE, 'stage', 'pending'),
##                              # 'pending' was set during this transaction:
##                              (VOLATILE, 'stage'),
##                              # More than one stage:
##                              (STATE, 'stage', None, 1),
##                              # Prior stage was not 'pending':
##                              (NOT, ('=', (STATE, 'stage', 'pending', 1),
##                                          (STATE, 'stage', None, 1))))),
##            (ACTIONS, (ABORT, ('Invalid stage transition:'
##                              ' non-pending to pending')))),

##        (   # Initial acknowledgement.
##            # Conditions: Pending and no reception alert set.
##            (CONDITION, (AND,
##                          (STATE, 'stage', 'pending'),
##                          (NOT, (STATE,
##                                 'alert', 'receipt ack', 0)))),
##            (ACTIONS,
##             (NOTIFY, 'Requester',
##              'Your %[var "aq_parent.title"]s tracker issue',
##              " '%[var title]s' (%[var id]s) has been received."
##              " "
##              "  Supporters have been notified, it can be viewed at"
##              " %[var absolute_url]s."),
##             (SET, 'alert', 'receipt ack'),
##             (SET, 'msgfromrole', 'requester'))),
        
##        (   # Ownership pending alerts.
##            (CONDITION, (AND,
##                          (VALUE, 'rsvps'),
##                          (STATE, 'stage', 'pending'),
##                          (OR,
##                           ('>', (STATE,), (VALUE, 'pending_due')),
##                           ('>', (STATE,), (VALUE, 'pending_overdue'))))),
##            # Due - either no alerts for notifications sent, or the last
##            # one was more than a day ago.
##            ( (CONDITION,
##               (OR, (NOT, (STATE, 'alert', 'owner due', 0)),
##                    ('>', (STATE, 'alert', 'owner due', 0),
##                          (VALUE, 'pending_due')))),
##              (ACTIONS,
####               (NOTIFY, 'Supporter',
####                "Issue %[var title]s - %[var id]s needs an owner!",
####                "Issue %[var issueIdentity]s has been pending more than %[var"
####                " pending_due]s days and needs an owner."),
##               (INCLUDE, 'due', 'ownership'),
##               (SET, 'alert', 'owner due'))),
##            (   # Overdue after more than pending_overdue days, and the overdue
##                # notice has not yet been sent.
##                (CONDITION,
##                 (AND,
##                  ('>', (STATE,), (VALUE, 'pending_overdue')),
##                  (NOT, (STATE,
##                         'alert', 'owner overdue', 0)))),
##                (ACTIONS,
####                 (NOTIFY, 'TrackerOwner',
####                  "Issue %[var title]s - %[var id]s is OVERDUE for ownership.",
####                  "Issue %[var issueIdentity]s is over %[var pending_overdue]"
####                  " old and needs an owner.  Supporters were alerted at"
####                  " the %[var pending_due] mark, but no dice."),
##                 (SET, 'alert', 'owner overdue'),
##                 (INCLUDE, 'overdue', 'ownership')))),

        (   # Clear up ownership pending state when suitable.
            (CONDITION, (NOT, (STATE, 'stage', 'pending'))),
            ( (CONDITION, (CONTAINS, 'due', 'ownership')),
              (ACTIONS, (REMOVE, 'due', 'ownership'))),
            ( (CONDITION, (CONTAINS, 'overdue', 'ownership')),
              (ACTIONS, (REMOVE, 'overdue', 'ownership')))),

        (   # Clear up rsvp pending state when suitable.
            (CONDITION, (AND, (OR, (CONTAINS, 'due', 'rsvp'),
                                   (CONTAINS, 'overdue', 'rsvp')),
                              (NOT, (STATE, 'msgfromrole', 'Requester', 0)))),
            ( (CONDITION, (CONTAINS, 'due', 'rsvp')),
              (ACTIONS, (REMOVE, 'due', 'rsvp'))),
            ( (CONDITION, (CONTAINS, 'overdue', 'rsvp')),
              (ACTIONS, (REMOVE, 'overdue', 'rsvp')))),

        (   # Alert to languishing reply.
            (CONDITION,
             # The last correspondence is from the requester:
             (AND,
              (OR, ('<',
                    (STATE, 'msgfromrole', ['Requester', 'TrackerOwner'], 0),
                    (STATE, 'msgfromrole', 'Supporter', 0)),
               ('>', (STATE, 'msgfromrole', 'Requester', 0), 3*ADAY)))),
             # Send daily alerts to issue owner, after initial grace period:
            ( (CONDITION, (OR,
                             (NOT, (STATE, 'alert', 'rsvp due', 0)),
                             ('>', (STATE, 'alert', 'rsvp due', 0), 1*ADAY))),
              (ACTIONS,
               (SET, 'alert', 'rsvp due'),
               (NOTIFY, 'IssueOwner',
                "Issue %[var title]s %[var id]s correspondence pending",
                "Issue %[var issueIdentity]s requester is owed"
                " correspondence.")
               )),
            # This has been going on too long - alert the tracker owner once:
            ( (CONDITION, (AND,
                            ('>', (STATE, 'alert', 'rsvp due', 0), 7*ADAY),
                            (NOT, (STATE, 'alert', 'rsvp overdue', 0)))),
              (ACTIONS,
               (SET, 'alert', 'rsvp overdue'),
               (NOTIFY, 'TrackerOwner',
                "Issue %[var title]s %[var id]s correspondence WAY pending",
                "Issue %[var issueIdentity]s requester has been owed"
                " correspondence for more than 7 days."))))
        ]
    return standard

