# General stuff ==========================================================
        
import string
import re
from DocumentTemplate import html_quote
import smtplib, socket
from DateTime import DateTime
import urllib

from AccessControl.PermissionRole import rolesForPermissionOn
from Products.ZCatalog.ZCatalog import Catalog

RELEVANCE_DEBUG = 1
RULES_DEBUG = 0
EMBLEM_DEBUG = 1

URLHEADexp = re.compile(r'(http|https|ftp|mailto)://[^ \t]+', re.IGNORECASE)

# Privacy stuff ==========================================================
        
# Role categories - map abstract tracker role names to tracker-specific zope
# security roles.  Use tracker_role() to do the translation from tracker dtml.
INVOLVED_ROLES = ['IssueOwner', 'Requester', 'TrackerOwner', 'Manager', 'Owner', 'ProjectMember']
STAFF_ROLES = ['TrackerOwner', 'IssueOwner', 'Supporter', 'Researcher', 'Manager', 'Owner', 'ProjectMember']
DEDICATED_ROLES = STAFF_ROLES + ['TrackerClient', 'Subscriber', 'Requester']
OPEN_ROLES = DEDICATED_ROLES + ['Anonymous', 'Member']
# XXX SupporterHelp would be a good place for exposition about the categories.
#     Some details would be good for TrackerElementsHelp, as well - see
#     set_privacy for related notes.
ROLE_CATEGORIES = {
    # staff: View any items, become issue owner, and do other state changes.
    #        Not shown other-correspondence items, but has access to them,
    #        including private ones.
    'staff': STAFF_ROLES,
    # observer: View correspondence beyond summary, private issues.  Does not 
    #           enable followup privileges or access to private items.
    'observer': (['Subscriber'] + INVOLVED_ROLES),
    # involved: Like observer but can submit followups.
    'involved': INVOLVED_ROLES,
    # dedicated: Can enter dedicated tracker.
    'dedicated': DEDICATED_ROLES,
    # open: Can enter open tracker.
    'open': OPEN_ROLES,
                   }

def elementEmblem(self, issue_id=None, item_id=None, nolink=0, complete=0,
                  forcelink=0, REQUEST=None):
    """Return a terse link specifying a particular issue, item, or tracker.

    If REQUEST.AUTHENTICATED_USER is not authorized to view the target element
    (or no REQUEST was passed in), then only the textual, unlinked version of
    the emblem is returned.  This serves as the cue that the element is
    unviewable, and prevents the user from trying to prevent the element and
    being "logged out".

    If optional 'complete' is 1, then items are indicated with both the
    issue and item parts.  If 'complete' is greater than 1, then the
    tracker abbreviation string is included.

    If optional nolink is true, then only the textual form of the emblem is
    returned.  This is useful when storing the emblem for future reference -
    elementEmblem can be applied then, obtaining the viewability info in the
    viewing context.

    Optional forcelink overrides both nolink and unviewability by the
    REQUESTer (or absence of the REQUEST altogether), forcing return of a
    linked version regardless of viewability.

    Unless specified, we fill in issue_id and item_id when we are within
    them, so you have to use aq_parent if you want the emblem for the parent.

    Generally, this routine is intended to make it easy to get the 'emblem'
    for the current element and for related ones - siblings, containers,
    contained, etc.  That's why it's not a method on each type of element."""

 ### XXX Eventually reconcile the logic to use get_elements().
##     tracker, issue, item = get_elements(self)
##     if issue_id: issue = None
##     if item_id: item = None

##    # DEBUG
##    import pdb
##    pdb.set_trace()

    if forcelink: nolink = 0

    candidate = self; ids = []
    tracker = issue = item = None
    # Go up the acquisition chain, gleaning the tracker and components.
    while 1:
        if not hasattr(candidate, 'IMATRACKER'):
            break
        if (hasattr(candidate, 'aq_explicit')
            and hasattr(candidate.aq_explicit, 'IMATRACKER')):
            tracker = candidate
            break
        if (not issue_id
            and hasattr(candidate, 'aq_explicit')
            and hasattr(candidate.aq_explicit, 'IMATRACKERISSUE')):
            issue, issue_id = candidate, candidate.id
        elif (not item_id
              and hasattr(candidate, 'aq_explicit')
              and hasattr(candidate.aq_explicit, 'IMATRACKERITEM')):
            item, item_id = candidate, candidate.id
        candidate = candidate.aq_parent

    if tracker is None:
        raise AttributeError, "Failed to glean tracker, %s" % html_quote(self)

    if (issue_id is None) and (item_id is None):
        got = tracker.title or tracker.id
        if nolink:
            return got
        else:
            return '<a href="%s">%s</a>' % (tracker.absolute_url(), got)

    if issue_id:
        issue = tracker[issue_id]

    issuegot = "(%s)" % issue_id
    if issue.private:
        issuegot = "(" + issuegot + ")"
        if not forcelink and not viewable(issue, REQUEST):
            nolink = 1
    if complete > 0:
        if item_id is None:
            issuegot = issuegot + "[]"
        if complete > 1:
            issuegot = tracker.abbrev + issuegot
    if item_id is None:                 # Emblem for an issue.
        if nolink: return issuegot
        else: return '<a href="%s">%s</a>' % (issue.absolute_url(), issuegot)
    else:                               # Emblem for an item.
        item = issue[item_id]
        itemgot = "[%s]" % item_id
        private = 0
        if item.private:
            private = 1
            itemgot = "[" + itemgot + "]"
        if ((private or not item_id in issue.promotes())
            and not forcelink and not viewable(item, REQUEST)):
            nolink = 1
        if complete: itemgot = issuegot + itemgot
        if nolink: return itemgot
        else: return '<a href="%s">%s</a>' % (item.absolute_url(), itemgot)

def tracker_role(self, roles, REQUEST=None, user=None):
    """Verify abstract tracker roles, in calling object, of REQUEST user.

    See ROLE_CATEGORIES "constant" for abstract roles.

    Special provisions are made for "anonymous" and "superuser" pseudo-roles - 
    the requesting user must be that user, not just have the roles.

    Anything else is taken as a concrete Zope role name to be checked for."""

    categories = ROLE_CATEGORIES
    candidates = []
    if type(roles) == type(''):
        roles = [roles]
    if not user:
        user = REQUEST.AUTHENTICATED_USER
    for i in roles:
        if (string.lower(i) in ["anonymous", "anonymous user"]
            and str(user.acl_users._nobody) == str(user)):
            return 1
        if (string.lower(i) == "superuser"
            and str(user.acl_users._super) == str(user)):
            return 1
        if not categories.has_key(i):
            candidates.append(i)
        else:
            candidates = candidates + categories[i]
    return user.has_role(candidates, self)

def issueOwners(self):
    """Return a list of the current issue's owner(s)."""
    if hasattr(self.aq_explicit, 'IMATRACKERISSUE'):
        issue = self
    else:
        tracker, issue, item = get_elements(self)
        if issue is None:
            raise 'Bad Request', ("No enclosing issue found - %s"
                                  % html_quote(`self`))
    got = []
    localroles = issue.__ac_local_roles__ or {}
    for who, roles in localroles.items():
        if 'IssueOwner' in roles:
            got.append(who)
    return got

def clear_local_role(role, obj, user=None):
    if user is not None:
        roleitems = ((user, obj.get_local_roles_for_userid(user)),)
    else:
        roleitems = (obj.__ac_local_roles__ or {}).items()
    for who, roles in roleitems:
        roles = filter(None, roles)
        newroles = list(roles)
        if role in newroles:
            newroles.remove(role)
            if newroles:
                obj.manage_setLocalRoles(who, newroles)
            else:
                # XXX This should remove entry for local user but just
                #     zeroes it?
                obj.manage_delLocalRoles([who])
            __traceback_info__ = ('reset', `roleitems`,
                                  `roles`, `newroles`, `obj`)

def get_acl_users_acquired(obj):
    """Return a list of all obj's acquired acl_users folders."""
    got = []
    had_acl_users = None
    while 1:
        if hasattr(obj, 'acl_users'):
            acl_users = obj.acl_users
            if acl_users is not had_acl_users:
                had_acl_users = acl_users
                got.append(acl_users)
        if hasattr(obj, 'aq_parent') and obj.aq_parent != obj:
            obj = obj.aq_parent
        else:
            break
    return got

def get_user_acquired(obj, userid):
    """Get a user obj in any of this object's acl_users folders, or None."""
    for au in get_acl_users_acquired(obj):
        got = au.getUser(userid)
        if got:
            return got
    return None

def verifyUser(self, userid, verified=None):
    # XXX we might should be ascending?
    if verified is None: verified = []
    if userid in verified:
        return 1
    else:
        got = get_user_acquired(self, userid)
        if got:
            verified.append(userid)
            return 1
        else:
            return 0

def viewable(object, REQUEST=None, userid=None):
    "True if REQUEST user has view access to object."
    if userid is not None:
        if userid == 'Anonymous User':
            # Doesn't matter which _nobody we use:
            userobj = object.acl_users._nobody
        else:
            # But we want to check everywhere for the real user:
            userobj = get_user_acquired(object, userid)
    else:
        userobj = (((REQUEST is not None)
                    and REQUEST.AUTHENTICATED_USER)
                   or None)
    if not userobj:
        return None
    roles = {}
    # (Slightly optimized in case there are a lot of 'needs' or 'roles'.)
    needs = rolesForPermissionOn('View', object)
    if 'Anonymous' in needs:
        return 1
    for r in userobj.getRolesInContext(object):
        roles[r] = None
    for i in needs:
        if roles.has_key(i):
            return 1
    return 0
##    ## Is has_permission broken??
##    return userobj.has_permission('View', object)

# Catalog and Searching ===================================================

def issuesSearch(self, features, prior=None, REQUEST=None):
    """Using a dictionary of features, refine a list of candidate elements.

    Features:

      type: 'strict', 'narrow', 'widen'
      any: keywords, matching any field
      requester: keywords
      private, public:  to *other* users - 'ON' or ()
      supporter: which supporter(s)
      stage: which stage(s) - []
      traits: {trait: [vals], ...} select traits with attr values.
      before (NYI): datestr - elements prior to specified date.
      after (NYI): datestr - elements subsequent to specified date.
      """

    from string import find

    searchtype = REQUEST.get('type', 'strict') or 'strict'
    widen = (searchtype == 'widen')
    if prior is None:
        priorobjs = []
        candidates = self.values()
    else:
        priorobjs = []
        for i in prior:
            __traceback_info__ = (`i`, `prior`, type(i))
            if type(i) == type(""):
                i = int(i)
            elif hasattr(i, 'id'):
                priorobjs.append(i)
                continue
            if self.has_key(i):
                priorobjs.append(self[i])
        if widen:
            candidates = self.values()
        elif searchtype == 'strict':
            candidates = self.values()
        else:
            candidates = priorobjs

    split, strip = string.split, string.strip

    anys = split(strip(features.get('any', '') or ''))

    stage = features.get('stage', [])
    supporters = features.get('supporter', [])
    requesters = split(strip(features.get('requester', '') or ''))
    private = features.get('private', 'ON') or 'ON'
    public = features.get('public', 'ON') or 'ON'
    traits = features.get('traits', {})

    got = []
    __traceback_info__ = (searchtype, 'num candidates', len(candidates))
    for e in candidates:
        if not viewable(e, REQUEST):
            continue
        qualified = 1
        curstage = e.stage
        if stage and (curstage not in stage) and (curstage not in anys):
            __traceback_info__ = __traceback_info__ + (`e`, "not " + `stage`)
            qualified = 0
        elif (e.private and not private) or (not e.private and not public):
            __traceback_info__ = __traceback_info__ + (`e`, 'failed private')
            qualified = 0
        elif supporters:
            qualified = 0
            for i in issueOwners(e):
                if i in supporters:
                    qualified = 1
                    break
                elif i in anys:
                    qualified = 1
                    break
            if not qualified:
                __traceback_info__ = (__traceback_info__ +
                                      (`e`, 'failed supporters', supporters))
        if qualified and requesters:
            currequester = e.requester
            if currequester not in requesters and currequester not in anys:
                __traceback_info__ = (__traceback_info__
                                      + (`e`, 'failed requesters'))
                qualified = 0
        if qualified and traits:
            for trait, val in e.traitVals:
                if val and val in anys:
                    break
                if traits.has_key(trait):
                    t = traits[trait]
                    if t and (val not in t):
                        qualified = 0
                        __traceback_info__ = (__traceback_info__
                                              + (`e`, 'failed traits',))
                        break
        if not qualified and anys:
            curtitle, curdescr = e.title, e.description
            for i in anys:
                if find(curtitle, i) != -1:
                    qualified = 1
                    break
                if find(curdescr, i) != -1:
                    qualified = 1
                    break
        if not qualified:
            __traceback_info__ = (__traceback_info__
                                  + (`e`, 'failed anys',))
            continue
        if priorobjs and widen:
            # Interject priors, if there are any, in order.
            curid = e.id
            while priorobjs and priorobjs[0].id < curid:
                got.append(priorobjs[0])
                del priorobjs[0]
            if priorobjs and priorobjs[0].id == curid:
                del priorobjs[0]
        got.append(e)
    if widen and not got:
        got = priorobjs
    return got

def lastVisible(seq, REQUEST):
    """Return the last element that is visible."""
##    # DEBUG
##    if REQUEST.AUTHENTICATED_USER.name == 'klmtest':
##        import pdb
##        pdb.set_trace()
    seq.reverse()
    for i in seq:
        if viewable(i, REQUEST):
            return i

def elementFilter(self, seq=None, refine='narrow',
                  attr=None, traits=None, negate=0, unpromoted=None,
                  visonly=0, REQUEST=None):
    """Return the elements with settings for attr in traits [XXX not to be
    confused with issue traits - this needs to be rectified], or not being
    traits if negate is specified.  If traits is none, just having the attr 
    qualifies (or disqualifies, if negated).

    If visonly is set, include only items accessible to the user
    (only after applying all the other criteria, including negate)."""

    widen = (refine == 'widen')
    if seq is None:
        seq = self.values()
    if widen:
        candidates = self.values()
    else:
        candidates = seq
    if unpromoted:
        promotes = self.promotes()

    got = []
    if type(traits) == type(""):
        traits = [traits]
    for i in candidates:
        qualifies = 0
        if not attr:
            qualifies = 1
        elif hasattr(i, attr):
            if not traits or (getattr(i, attr) in traits):
                qualifies = 1
        if qualifies and unpromoted and (int(i.id) in promotes):
            qualifies = 0
        if (qualifies and not negate) or (negate and not qualifies):
            if (not visonly or viewable(i, REQUEST)):
                if seq and widen:
                    # Include prior elements from the initial set first.
                    curid = i.id
                    while seq and seq[0].id < curid:
                        got.append(seq[0])
                        del seq[0]
                    if seq and seq[0].id == curid:
                        del seq[0]
                got.append(i)
    if widen and not got:
        got = seq
    return got

# Miscellaneous element data interfaces. ================================

def tracker_data_version_actions(tracker):
    if not hasattr(tracker.aq_explicit, 'DATA_VERSION'):
        tracker.aq_explicit.DATA_VERSION = TRACKER_DATA_VERSION
        return
    if tracker.aq_explicit.DATA_VERSION != TRACKER_DATA_VERSION:
        # Activities for migration to TRACKER_DATA_VERSION 1:
        tracker.aq_explicit.DATA_VERSION = TRACKER_DATA_VERSION

def issue_data_version_actions(issue):
    if not hasattr(issue.aq_explicit, 'DATA_VERSION'):
        issue.aq_explicit.DATA_VERSION = ISSUE_DATA_VERSION
        return
    if issue.aq_explicit.DATA_VERSION != ISSUE_DATA_VERSION:
        # Activities for migration to ISSUE_DATA_VERSION 1:
        issue.aq_explicit.DATA_VERSION = ISSUE_DATA_VERSION

def item_data_version_actions(item):
    if not hasattr(item.aq_explicit, 'DATA_VERSION'):
        item.aq_explicit.DATA_VERSION = ITEM_DATA_VERSION
        return
    if item.aq_explicit.DATA_VERSION != ITEM_DATA_VERSION:
        # Activities for migration to ITEM_DATA_VERSION 1:
        item.aq_explicit.DATA_VERSION = ITEM_DATA_VERSION

# Sendmail stuff. ================================================

def relevant_supporters(item, issue=None, tracker=None):
    """Identify candidate supporters according to request traits."""
    if not ((tracker is not None) and (issue is not None)):
        t, i, x = get_elements(item)
        if tracker is None: tracker = t
        if issue is None: issue = i

    traitVals = issue.traitVals
    got = []

    for supporter, assignments in issue.assignments.items():
        disqualified = 0

        for trait, val in traitVals:
            if assignments.has_key(trait):
                # Qualifying trait, _if_ there are selected vals for it.
                selected = assignments[trait]
                if not selected:
                    # No selections means insensitive to this trait.
                    continue
                invert = (selected[0] == ())
                if invert:
                    selected = selected[1:]
                if val in selected:
                    disqualified = invert
                else:
                    disqualified = not invert

            if disqualified:
                break
        if not disqualified and assignments.has_key(" __requesters"):
            selected = assignments[" __requesters"]
            if selected:
                invert = (selected[0] == ())
                if invert:
                    selected = selected[1:]
                if issue.requester in selected:
                    disqualified = invert

        if not disqualified:
            got.append(supporter)
    
    return got

def send_email(tracker, sender, destinations, message):
    """Send a fully composed message to indicated destinations."""
    if destinations:
        server = smtplib.SMTP(tracker.smtphost, tracker.smtpport)
        server.sendmail(sender, destinations, message)
        server.quit()

# Miscellaneous tools ================================================

def get_elements(element):
    """Return the tracker, issue, and item determined by a tracker element.

    Ie, tracker, issue, and item for an item, tracker and issue for an issue,
    and just the tracker for a tracker element.  'None' is returned for any
    indeterminate elements."""
    if element.type == 'tracker_item':
        return (element.aq_parent.aq_parent, element.aq_parent, element)
    elif element.type == 'tracker_issue':
        return (element.aq_parent, element, None)
    elif element.type == 'tracker':
        return (element, None, None)
    else:
        raise ValueError, "%s not a tracker element." % html_quote(element)

def XXXsuccinct_time(t):
    # XXX It's too bad - for some reason, passing the date across the dtml
    #     boundary is slooow.  Doing (less elaborate) logic in dtml, instead.
    """Return representation of a date/time tailored w.r.t. the current time.

     - if within the last day, give the am/pm time
     - or if within the current year, give the date/time sans year
     - else give the whole shebang."""

    if type(t) == type("") or type(t) == type(1.0):
        t = DateTime(t)
    if t.isCurrentDay():
        return t.AMPMMinutes()
    else:
        yr = ""
        if t.year() != DateTime().year():
            yr = " " + str(t.year())
        return "%s %s%s %s" % (t.aMonth(), t.day(), yr, t.AMPMMinutes())

def diffs(l1, l2):
    """Return two lists - elements in l1 not in l2, and vice versa.
    Preserves order."""
    l1u, l1d, l2u, l2d = [], {}, [], {}
    for i in l1: l1d[i] = 0
    for i in l2: l2d[i] = 0
    for i in l1:
        if (not l2d.has_key(i)): l1u.append(i)
    for i in l2:
        if (not l1d.has_key(i)): l2u.append(i)
    return (l1u, l2u)
def unique(lst, noblanks=0):
    """Return a copy of list with all repeats removed.  Preserves order."""
    gotd = {}
    for i in lst:
        if noblanks and i is '': continue
        if not gotd.has_key(i): gotd[i] = 0
    got = []
    for i in lst:
        if gotd.has_key(i):
            got.append(i)
            del gotd[i]
    return got

def hrefize(self, text, emblem_url_only=0, forcelink=0, complete=1,
            REQUEST=None,
            find=string.find, rfind=string.rfind, strip=string.strip,
            atoi=string.atoi):
    """Recognize absolute URLs and element emblems and turn them into links.

    We return (newtext, [references]), where each reference is either the
    indicated tracker element object or a URL string, if not an element."""
    tracker, issue, item = get_elements(self)
    gotlines, gotrefs = [], []
    lines = string.split(text, '\n')
    URLsearch = URLHEADexp.search
    for line in lines:
        stripped_br = 0
        line = strip(line)
        linelen = len(line)

        if not emblem_url_only:
            # Glean absolute URLs:
            end = 0
            while 1:
                found = URLsearch(line, pos=end)
                if not found:
                    break
                url = found.group()
                end = found.end()
                after = len(line) - end
                line = (line[0:found.start()]
                        + ("<a href=\"%s\">%s</a>" % (url, url))
                        + line[end:])
                gotrefs.append(url)
                # Account for change in line length!
                end = len(line) - after
                if not after:
                    break

        # Glean element emblems:
        begin = len(line)
        emblem = ''
        while 1:
            begin = rfind(line, ")[", 0, begin)
            if begin == -1:
                break
            itemend = find(line, ']', begin)
            if itemend == -1:
                continue
            itemnum = strip(line[begin+2:itemend]) or None
            if itemnum and (itemnum[0] == '['):
                itemnum = itemnum[1:]
                itemend = itemend + 1
            issuebegin = rfind(line, "(", 0, begin)
            if issuebegin == -1:
                continue
            issuenum = strip(line[issuebegin+1:begin]) or None
            if issuenum and (issuenum[-1] == ')'):
                issuenum = issuenum[:-1]
                issuebegin = issuebegin - 1
            if not issuenum and not itemnum:
                continue
            try:
                if itemnum: itemnum = atoi(itemnum)
                if itemnum == 0: itemnum = None
                if issuenum: issuenum = atoi(issuenum)
                if not (issuenum or itemnum):
                    raise ValueError    # Continue at the following except.
                theissue = issue
                if issuenum: theissue = tracker[issuenum]
                if itemnum:
                    element = theissue[itemnum]
                else:
                    element = theissue
                emblem = elementEmblem(element, complete=complete,
                                       forcelink=forcelink, REQUEST=REQUEST)
            except (KeyError, ValueError), msg:
                continue
            if emblem:
                gotrefs.append(element)
                line = (line[0:issuebegin] + emblem + line[itemend+1:])
                emblem = ''
        if stripped_br:
            line = line + "<br>"
        gotlines.append(line)
    return (string.join(gotlines, '\n'), gotrefs)


def html_quote_indented(self, text, maxchars=None, maxlines=None,
                        do_hrefize=0, 
                        etc='...', charsbreak='', linesbreak='',
                        find=string.find, rfind=string.rfind,
                        REQUEST=None):
    """Like html_quote, + preserve indentation and don't wrap short lines."""
    from ts_regex import match
    from string import strip, split, expandtabs, atoi
    lines = split(html_quote(expandtabs(text, 8)), '\n')
    while lines and not strip(lines[-1]):
        del lines[-1]
    got = []
    for line in lines:
        line = line + " <br>"
        if line[0] == ' ':
            pref = match(" +", line)
            got.append((pref * "&nbsp;") + line[pref:])
        else:
            got.append(line)
    while got and (not string.strip(got[-1]) or got[-1] == '<p>'):
        del got[-1]
    origlines = len(got)
    charstruncated = linestruncated = 0
    if maxlines and len(got) > maxlines:
        got = got[:maxlines]
        linestruncated = 1
    newtext = string.join(got, '\n')
    if maxchars and (len(newtext) > maxchars):
        newtext = newtext[:maxchars]
        charstruncated = 1
    if charstruncated:
        newtext = newtext + charsbreak + etc
    elif linestruncated:
        newtext = newtext + linesbreak + etc
    if do_hrefize:
        newtext = hrefize(self, newtext, REQUEST=REQUEST)[0]
    return newtext

def double_line(text, wrapbefore):
    """Return text copy limited to 2 x wrapbefore, and wrapped with <br>.

    Return text if it's not longer than 'wrapbefore'.

    Result is html_quote'ed"""
    (first, second) = split_before_whitespace(text, wrapbefore)
    if not second:
        return html_quote(first[:wrapbefore])
    if len(second) > wrapbefore:
        (second, discard) = split_before_whitespace(second, wrapbefore)
    if len(second) > wrapbefore:
        second = second[:wrapbefore]
    return html_quote(first) + "<br>" + html_quote(second)
                
def split_before_whitespace(text, splitbefore):
    "Return (first, second) split at recent whitespace before splitbefore."
    prior = splitbefore
    if len(text) <= splitbefore:
        return (text, '')
    while prior > 0 and text[prior] not in string.whitespace:
        prior = prior - 1
    if not prior:
        return (text, '')
    else:
        return (text[:prior], text[prior+1:])
    
def pedigree(self):
    """Investigation utility showing the base class hierachy of a class."""
    __traceback_info__ = (`self`)
    structure = pedigree_structure(self.__class__)
    got = indented(structure, depth=0)
    return ("<!--#var standard_html_header-->"
            + ("<h3>%s Instance Inherited Classes</h3>\n" % got[0])
            + "\n<em>('...' abbreviates already elaborated classes)</em>"
            + "\n<pre>\n" + string.join(got[1:], '\n') + "\n</pre>\n"
            + "<!--#var standard_html_footer-->")
    
def indented(nesting, depth=0):
    """Indicate list nesting with indentation instead of containment.

    Ellipses appreviate already elaborated classes, and single-element
    inheritance is shown on the same line as the inheritor."""
    got = []
    for i in nesting:
        if type(i) == type(''):
            got.append((depth * '  ') + i)
##         elif len(i) == 2 and len(i[1]) == 1:
##             got.append((depth+1) * '  ' + "%s %s" % (i[0], i[1][0]))
        else:
            got = got + indented(i, depth + 1)
    return got

def pedigree_structure(klass, already_seen=None):
    """Return nested lists reflecting inheritance structure of self."""
    __traceback_info__ = (`klass`)
    bases = klass.__bases__
    seen = ""
    if already_seen is None:
        already_seen = []
    elif bases and klass in already_seen:
        seen = " ..."

    if hasattr(klass, '__module__') and klass.__module__:
        got = ["%s (%s)%s" % (klass.__name__, klass.__module__, seen)]
    else:
        got = [klass.__name__ + seen]

    if not seen:
        already_seen.append(klass)
        for i in bases:
            got.append(pedigree_structure(i, already_seen))

    return got
